# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Any

import gt4py.next.ffront.field_operator_ast as foast
from gt4py.eve import NodeTranslator, traits
from gt4py.next import errors
from gt4py.next.type_system import type_translation


@dataclass
class ClosureVarFolding(NodeTranslator, traits.VisitorWithSymbolTableTrait):
    """
    Replace references to closure variables or their attributes with constants.

    `Name` nodes that refer to closure variables are replaced by `Constant`
     nodes. `Attribute` nodes that refer to attributes of closure variables
     are recursively replaced by `Constant` nodes.
    """

    closure_vars: dict[str, Any]

    @classmethod
    def apply(
        cls, node: foast.FunctionDefinition | foast.FieldOperator, closure_vars: dict[str, Any]
    ) -> foast.FunctionDefinition:
        return cls(closure_vars=closure_vars).visit(node)

    def visit_Name(
        self,
        node: foast.Name,
        current_closure_vars: dict[str, Any],
        symtable: dict[str, foast.Symbol],
        **kwargs: Any,
    ) -> foast.Name | foast.Constant:
        if node.id in symtable:
            definition = symtable[node.id]
            if definition in current_closure_vars:
                value = self.closure_vars[node.id]
                if isinstance(value, type_translation.ConstantPythonNamespaceObject):
                    return foast.Constant(value=value, location=node.location)
        return node

    def visit_Attribute(
        self, node: foast.Attribute, **kwargs: Any
    ) -> foast.Constant | foast.Attribute:
        value = self.visit(node.value, **kwargs)
        if isinstance(value, foast.Constant):
            if hasattr(value.value, node.attr):
                return foast.Constant(value=getattr(value.value, node.attr), location=node.location)
            raise errors.MissingAttributeError(node.location, node.attr)
        return node

    def visit_FunctionDefinition(
        self, node: foast.FunctionDefinition, **kwargs: Any
    ) -> foast.FunctionDefinition:
        return self.generic_visit(node, current_closure_vars=node.closure_vars, **kwargs)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
from typing import Any, Optional

from gt4py.next.ffront import (
    dialect_ast_enums,
    foast_to_gtir,
    program_ast as past,
    stages as ffront_stages,
    type_info as ffront_type_info,
    type_specifications as ts_ffront,
)
from gt4py.next.ffront.past_passes import closure_var_type_deduction, type_deduction
from gt4py.next.ffront.stages import ConcreteFOASTOperatorDef, ConcretePASTProgramDef
from gt4py.next.iterator import ir as itir
from gt4py.next.otf import toolchain, workflow
from gt4py.next.type_system import type_info, type_specifications as ts


@dataclasses.dataclass(frozen=True)
class ItirShim:
    """
    A wrapper for a FOAST operator definition with `__gt_*__` special methods.

    Can be placed in a PAST program definition's closure variables so the program
    lowering has access to the relevant information.
    """

    definition: ConcreteFOASTOperatorDef
    foast_to_itir: workflow.Workflow[ConcreteFOASTOperatorDef, itir.FunctionDefinition]

    def __gt_closure_vars__(self) -> Optional[dict[str, Any]]:
        return self.definition.data.closure_vars

    def __gt_type__(self) -> ts.CallableType:
        assert isinstance(self.definition.data.foast_node.type, ts.CallableType)
        return self.definition.data.foast_node.type

    def __gt_itir__(self) -> itir.FunctionDefinition:
        return self.foast_to_itir(self.definition)

    # FIXME[#1582](tehrengruber): remove after refactoring to GTIR
    def __gt_gtir__(self) -> itir.FunctionDefinition:
        # backend should have self.foast_to_itir set to foast_to_gtir
        return self.foast_to_itir(self.definition)


@dataclasses.dataclass(frozen=True)
class OperatorToProgram(workflow.Workflow[ConcreteFOASTOperatorDef, ConcretePASTProgramDef]):
    """
    Generate a PAST program definition from a FOAST operator definition.

    This workflow step must must be given a FOAST -> ITIR lowering step so that it can place
    valid `ItirShim` instances into the closure variables of the generated program.

    Example:
        >>> from gt4py import next as gtx
        >>> from gt4py.next.otf import arguments, toolchain
        >>> IDim = gtx.Dimension("I")

        >>> @gtx.field_operator
        ... def copy(a: gtx.Field[[IDim], gtx.float32]) -> gtx.Field[[IDim], gtx.float32]:
        ...     return a

        >>> op_to_prog = OperatorToProgram(foast_to_gtir.adapted_foast_to_gtir_factory())

        >>> compile_time_args = arguments.CompileTimeArgs(
        ...     args=(
        ...         *(param.type for param in copy.foast_stage.foast_node.definition.params),
        ...         copy.foast_stage.foast_node.definition.type.returns,
        ...     ),
        ...     kwargs={},
        ...     offset_provider={"I": IDim},
        ...     column_axis=None,
        ...     argument_descriptor_contexts={},
        ... )

        >>> copy_program = op_to_prog(
        ...     toolchain.ConcreteArtifact(copy.foast_stage, compile_time_args)
        ... )

        >>> print(copy_program.data.past_node.id)
        __field_operator_copy

        >>> assert copy_program.data.closure_vars["copy"].definition.data is copy.foast_stage
    """

    foast_to_itir: workflow.Workflow[ConcreteFOASTOperatorDef, itir.FunctionDefinition]

    def __call__(self, inp: ConcreteFOASTOperatorDef) -> ConcretePASTProgramDef:
        # TODO(tehrengruber): implement mechanism to deduce default values
        #  of arg and kwarg types
        # TODO(tehrengruber): check foast operator has no out argument that clashes
        #  with the out argument of the program we generate here.
        arg_types, kwarg_types = inp.args.args, inp.args.kwargs
        assert not kwarg_types

        type_ = inp.data.foast_node.type
        loc = inp.data.foast_node.location
        assert isinstance(inp.data.foast_node.type, ts.CallableType)
        partial_program_type = ffront_type_info.type_in_program_context(inp.data.foast_node.type)
        assert isinstance(partial_program_type, ts_ffront.ProgramType)
        args_names = [
            *partial_program_type.definition.pos_only_args,
            *partial_program_type.definition.pos_or_kw_args.keys(),
            *partial_program_type.definition.kw_only_args.keys(),
        ]
        assert isinstance(type_, ts.CallableType)
        assert arg_types[-1] == type_info.return_type(
            type_, with_args=list(arg_types), with_kwargs=kwarg_types
        )
        assert args_names[-1] == "out"

        params_decl: list[past.Symbol] = [
            past.DataSymbol(
                id=name,
                type=arg_type,
                namespace=dialect_ast_enums.Namespace.LOCAL,
                location=loc,
            )
            for name, arg_type in zip(
                args_names,
                arg_types,
                strict=True,
            )
        ]
        params_ref = [past.Name(id=pdecl.id, location=loc) for pdecl in params_decl[:-1]]
        out_ref = past.Name(id="out", location=loc)

        if inp.data.foast_node.id in inp.data.closure_vars:
            raise RuntimeError("A closure variable has the same name as the field operator itself.")

        closure_symbols: list[past.Symbol] = [
            past.Symbol(
                id=inp.data.foast_node.id,
                type=ts.DeferredType(constraint=None),
                namespace=dialect_ast_enums.Namespace.CLOSURE,
                location=loc,
            ),
        ]

        fieldop_itir_closure_vars = {inp.data.foast_node.id: ItirShim(inp, self.foast_to_itir)}

        untyped_past_node = past.Program(
            id=f"__field_operator_{inp.data.foast_node.id}",
            type=ts.DeferredType(constraint=ts_ffront.ProgramType),
            params=params_decl,
            body=[
                past.Call(
                    func=past.Name(id=inp.data.foast_node.id, location=loc),
                    args=params_ref,
                    kwargs={"out": out_ref},
                    location=loc,
                )
            ],
            closure_vars=closure_symbols,
            location=loc,
        )
        untyped_past_node = closure_var_type_deduction.ClosureVarTypeDeduction.apply(
            untyped_past_node,
            fieldop_itir_closure_vars,  # type: ignore[arg-type]
        )
        past_node = type_deduction.ProgramTypeDeduction.apply(untyped_past_node)

        return toolchain.ConcreteArtifact(
            data=ffront_stages.PASTProgramDef(
                past_node=past_node,
                closure_vars=fieldop_itir_closure_vars,  # type: ignore[arg-type]
                grid_type=inp.data.grid_type,
            ),
            args=inp.args,
        )


def operator_to_program_factory(
    foast_to_itir_step: Optional[
        workflow.Workflow[ConcreteFOASTOperatorDef, itir.FunctionDefinition]
    ] = None,
    cached: bool = True,
) -> workflow.Workflow[ConcreteFOASTOperatorDef, ConcretePASTProgramDef]:
    """Optionally wrap `OperatorToProgram` in a `CachedStep`."""
    wf: workflow.Workflow[ConcreteFOASTOperatorDef, ConcretePASTProgramDef] = OperatorToProgram(
        foast_to_itir_step or foast_to_gtir.adapted_foast_to_gtir_factory()
    )
    if cached:
        wf = workflow.CachedStep(wf, hash_function=ffront_stages.fingerprint_stage)
    return wf



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import re
from functools import lru_cache

import numpy as np
from dace import data, dtypes, symbolic

from gt4py import eve
from gt4py.cartesian.gtc import common
from gt4py.storage.cartesian import layout


def get_dace_debuginfo(node: common.LocNode) -> dtypes.DebugInfo:
    if node.loc is None:
        return dtypes.DebugInfo(0)

    return dtypes.DebugInfo(
        node.loc.line, node.loc.column, node.loc.line, node.loc.column, node.loc.filename
    )


def array_dimensions(array: data.Array) -> list[bool]:
    return [
        any(
            re.match(f"__.*_{k}_stride", str(sym))
            for st in array.strides
            for sym in symbolic.pystr_to_symbolic(st).free_symbols
        )
        or any(
            re.match(f"__{k}", str(sym))
            for sh in array.shape
            for sym in symbolic.pystr_to_symbolic(sh).free_symbols
        )
        for k in "IJK"
    ]


def replace_strides(arrays: list[data.Array], layout_info: layout.LayoutInfo) -> dict[str, str]:
    symbol_mapping = {}
    for array in arrays:
        dims = array_dimensions(array)
        ndata_dims = len(array.shape) - sum(dims)
        axes = [ax for ax, m in zip("IJK", dims) if m] + [str(i) for i in range(ndata_dims)]
        layout = layout_info["layout_map"](tuple(axes))

        stride = 1
        for idx in reversed(np.argsort(layout)):
            symbol = array.strides[idx]
            if symbol.is_symbol:
                symbol_mapping[str(symbol)] = symbolic.pystr_to_symbolic(stride)
            stride *= array.shape[idx]

    return symbol_mapping


def data_type_to_dace_typeclass(data_type: common.DataType) -> dtypes.typeclass:
    dtype = np.dtype(common.data_type_to_typestr(data_type))
    return dtypes.typeclass(dtype.type)


@lru_cache(maxsize=None)
def get_dace_symbol(
    name: eve.SymbolRef, dtype: common.DataType = common.DataType.INT32
) -> symbolic.symbol:
    return symbolic.symbol(name, dtype=data_type_to_dace_typeclass(dtype))



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py import eve
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, ir_makers as im


class CollapseListGet(eve.PreserveLocationVisitor, eve.NodeTranslator):
    """Simplifies expressions containing `list_get`.

    Examples
      - `list_get(i, neighbors(o, it))` -> `deref(shift(o, i)(it))`
      - `list_get(i, make_const_list(e))` -> `e`
    """

    def visit_FunCall(self, node: itir.FunCall, **kwargs) -> itir.Node:
        node = self.generic_visit(node)
        if cpm.is_call_to(node, "list_get"):
            if cpm.is_call_to(node.args[1], "if_"):
                list_idx = node.args[0]
                cond, true_val, false_val = node.args[1].args
                return im.if_(
                    cond,
                    self.visit(im.list_get(list_idx, true_val)),
                    self.visit(im.list_get(list_idx, false_val)),
                )
            if cpm.is_call_to(node.args[1], "neighbors"):
                offset_tag = node.args[1].args[0]
                offset_index = (
                    itir.OffsetLiteral(value=int(node.args[0].value))
                    if isinstance(node.args[0], itir.Literal)
                    else node.args[
                        0
                    ]  # else-branch: e.g. SymRef from unroll_reduce, TODO(havogt): remove when we replace unroll_reduce by list support in gtfn
                )
                it = node.args[1].args[1]
                return im.deref(im.shift(offset_tag, offset_index)(it))
            if cpm.is_call_to(node.args[1], "make_const_list"):
                return node.args[1].args[0]
            if cpm.is_applied_map(node.args[1]):
                # list_get(0, map_(λ(val_) → foo(val_, int64))(·__sym_1))
                # -> (λ(val_) → foo(val_, int64))(list_get(0, ·__sym_1))
                lsts = node.args[1].args
                assert len(node.args[1].fun.args) == 1  # a single lambda in the map
                mapped_lambda = node.args[1].fun.args[0]
                res = im.call(mapped_lambda)(*[im.list_get(node.args[0], lst) for lst in lsts])
                return res

        return node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from dace.sdfg.analysis.schedule_tree import treenodes as tn

from gt4py.cartesian.gtc.dace.treeir import Axis


class SwapHorizontalMaps(tn.ScheduleNodeVisitor):
    """
    Given a schedule tree with K-JI loops, swap the IJ for a K-IJ loop.

    Expected input is something like

    // horizontal loop - potentially nested
    map __j, __i in [0:J, 0:I]
        // computation here


    Expected output swaps the indices & range

    // horizontal loop - potentially nested
    map __i, __j in [0:I, 0:J]
        // computation here
    """

    def visit_MapScope(self, node: tn.MapScope):
        if node.node.params[0].startswith(Axis.J.iteration_symbol()) and node.node.params[
            1
        ].startswith(Axis.I.iteration_symbol()):
            # Swap params
            tmp_index = node.node.params[0]
            node.node.params[0] = node.node.params[1]
            node.node.params[1] = tmp_index
            # Swap ranges
            tmp_bounds = node.node.range[0]
            node.node.range[0] = node.node.range[1]
            node.node.range[1] = tmp_bounds

        self.visit(node.children)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import Dict, Final, Mapping, Optional, TypeVar

import dace

from gt4py import eve
from gt4py.eve.extended_typing import NestedTuple
from gt4py.next import common as gtx_common, utils as gtx_utils
from gt4py.next.iterator import ir as gtir
from gt4py.next.iterator.ir_utils import ir_makers as im
from gt4py.next.program_processors.runners.dace.lowering import gtir_python_codegen
from gt4py.next.type_system import type_specifications as ts


_TASKLET_CONNECTOR_PREFIX: Final[str] = "__tlet_"
"""Prefix string to be used for tasklet connectors."""


def debug_info(
    node: gtir.Node, *, default: Optional[dace.dtypes.DebugInfo] = None
) -> Optional[dace.dtypes.DebugInfo]:
    """Include the GT4Py node location as debug information in the corresponding SDFG nodes."""
    location = node.location
    if location:
        return dace.dtypes.DebugInfo(
            start_line=location.line,
            start_column=location.column if location.column else 0,
            end_line=location.end_line if location.end_line else -1,
            end_column=location.end_column if location.end_column else 0,
            filename=location.filename,
        )
    return default


def get_map_variable(dim: gtx_common.Dimension) -> str:
    """
    Format map variable name based on the naming convention for application-specific SDFG transformations.
    """
    suffix = "dim" if dim.kind == gtx_common.DimensionKind.LOCAL else ""
    return f"i_{dim.value}_gtx_{dim.kind}{suffix}"


def make_tasklet_connector_for(name: str) -> str:
    """
    Format tasklet connector name based on the naming convention to avoid conflicts with GTIR program symbols.
    """
    assert _TASKLET_CONNECTOR_PREFIX.startswith("__")
    assert not name.startswith(_TASKLET_CONNECTOR_PREFIX)
    return f"{_TASKLET_CONNECTOR_PREFIX}{name.lstrip('_')}"


def make_symbol_tree(tuple_name: str, tuple_type: ts.TupleType) -> NestedTuple[gtir.Sym]:
    """
    Creates a tree representation of the symbols corresponding to the tuple fields.
    The constructed tree preserves the nested nature of the tuple type, if any.

    Examples
    --------
    >>> sty = ts.ScalarType(kind=ts.ScalarKind.INT32)
    >>> fty = ts.FieldType(dims=[], dtype=ts.ScalarType(kind=ts.ScalarKind.FLOAT32))
    >>> t = ts.TupleType(types=[sty, ts.TupleType(types=[fty, sty])])
    >>> assert make_symbol_tree("a", t) == (
    ...     im.sym("a_0", sty),
    ...     (im.sym("a_1_0", fty), im.sym("a_1_1", sty)),
    ... )
    """
    assert all(isinstance(t, ts.DataType) for t in tuple_type.types)
    fields = [(f"{tuple_name}_{i}", field_type) for i, field_type in enumerate(tuple_type.types)]
    return tuple(
        make_symbol_tree(field_name, field_type)
        if isinstance(field_type, ts.TupleType)
        else im.sym(field_name, field_type)
        for field_name, field_type in fields
    )


def flatten_tuple_fields(tuple_name: str, tuple_type: ts.TupleType) -> list[gtir.Sym]:
    """
    Creates a list of symbols, annotated with the data type, for all elements of the given tuple.

    Examples
    --------
    >>> sty = ts.ScalarType(kind=ts.ScalarKind.INT32)
    >>> fty = ts.FieldType(dims=[], dtype=ts.ScalarType(kind=ts.ScalarKind.FLOAT32))
    >>> t = ts.TupleType(types=[sty, ts.TupleType(types=[fty, sty])])
    >>> assert flatten_tuple_fields("a", t) == [
    ...     im.sym("a_0", sty),
    ...     im.sym("a_1_0", fty),
    ...     im.sym("a_1_1", sty),
    ... ]
    """
    symbol_tree = make_symbol_tree(tuple_name, tuple_type)
    return list(gtx_utils.flatten_nested_tuple(symbol_tree))


def replace_invalid_symbols(ir: gtir.Program) -> gtir.Program:
    """
    Ensure that all symbols used in the program IR are valid strings (e.g. no unicode-strings).

    If any invalid symbol present, this function returns a copy of the input IR where
    the invalid symbols have been replaced with new names. If all symbols are valid,
    the input IR is returned without copying it.
    """

    class ReplaceSymbols(eve.PreserveLocationVisitor, eve.NodeTranslator):
        PRESERVED_ANNEX_ATTRS = ("domain",)

        T = TypeVar("T", gtir.Sym, gtir.SymRef)

        def _replace_sym(self, node: T, symtable: Dict[str, str]) -> T:
            sym = str(node.id)
            return type(node)(id=symtable.get(sym, sym), type=node.type)

        def visit_Sym(self, node: gtir.Sym, *, symtable: Dict[str, str]) -> gtir.Sym:
            return self._replace_sym(node, symtable)

        def visit_SymRef(self, node: gtir.SymRef, *, symtable: Dict[str, str]) -> gtir.SymRef:
            return self._replace_sym(node, symtable)

    # program arguments are checked separetely, because they cannot be replaced
    if not all(dace.dtypes.validate_name(str(sym.id)) for sym in ir.params):
        raise ValueError("Invalid symbol in program parameters.")

    # check there are no conflicts with the connctor names generated in the lowering
    if any(str(sym.id).startswith(_TASKLET_CONNECTOR_PREFIX) for sym in ir.params):
        raise ValueError(
            f"Unexpectd symbol with prefix '{_TASKLET_CONNECTOR_PREFIX}' in program parameters."
        )

    ir_sym_ids = {str(sym.id) for sym in eve.walk_values(ir).if_isinstance(gtir.Sym).to_set()}
    ir_ssa_uuid = eve.utils.SequentialIDGenerator(prefix="gtir_var")

    # note: traverse in alphabetical order to generate UIDs in deterministic way
    invalid_symbols_mapping = {
        sym_id: next(ir_ssa_uuid)
        for sym_id in sorted(ir_sym_ids)
        if not dace.dtypes.validate_name(sym_id)
    }
    if len(invalid_symbols_mapping) == 0:
        return ir

    # assert that the new symbol names are not used in the IR
    assert ir_sym_ids.isdisjoint(invalid_symbols_mapping.values())
    return ReplaceSymbols().visit(ir, symtable=invalid_symbols_mapping)


def get_symbolic(ir: gtir.Expr) -> dace.symbolic.SymbolicType:
    """
    Specialized visit method for symbolic expressions.

    Returns:
        A dace symbolic expression of the given GTIR.
    """
    python_source = gtir_python_codegen.get_source(ir)
    return dace.symbolic.pystr_to_symbolic(python_source)


def safe_replace_symbolic(
    val: dace.symbolic.SymbolicType,
    symbol_mapping: Mapping[dace.symbolic.SymbolicType | str, dace.symbolic.SymbolicType | str],
) -> dace.symbolic.SymbolicType:
    """
    Replace free symbols in a dace symbolic expression, using `safe_replace()`
    in order to avoid clashes in case the new symbol value is also a free symbol
    in the original exoression.

    Args:
        val: The symbolic expression where to apply the replacement.
        symbol_mapping: The mapping table for symbol replacement.

    Returns:
        A new symbolic expression as result of symbol replacement.
    """
    # The list `x` is needed because `subs()` returns a new object and can not handle
    # replacement dicts of the form `{'x': 'y', 'y': 'x'}`.
    # The utility `safe_replace()` will call `subs()` twice in case of such dicts.
    x = [val]
    dace.symbolic.safe_replace(symbol_mapping, lambda m, xx=x: xx.append(xx[-1].subs(m)))
    return x[-1]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses

from gt4py.next.otf import definitions, stages, workflow


@dataclasses.dataclass(frozen=True)
class OTFCompileWorkflow(workflow.NamedStepSequence):
    """The typical compiled backend steps composed into a workflow."""

    translation: definitions.TranslationStep
    bindings: workflow.Workflow[stages.ProgramSource, stages.CompilableProject]
    compilation: workflow.Workflow[stages.CompilableProject, stages.ExecutableProgram]
    decoration: workflow.Workflow[stages.ExecutableProgram, stages.ExecutableProgram]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import typing
from typing import Generic

from gt4py.next.otf import workflow


S = typing.TypeVar("S")
T = typing.TypeVar("T")
DefT = typing.TypeVar("DefT")
ArgsT = typing.TypeVar("ArgsT")


@dataclasses.dataclass
class ConcreteArtifact(Generic[DefT, ArgsT]):
    data: DefT
    args: ArgsT


@dataclasses.dataclass(frozen=True)
class DataOnlyAdapter(
    workflow.ChainableWorkflowMixin,
    workflow.ReplaceEnabledWorkflowMixin,
    workflow.Workflow[ConcreteArtifact[S, ArgsT], ConcreteArtifact[T, ArgsT]],
    Generic[ArgsT, S, T],
):
    step: workflow.Workflow[S, T]

    def __call__(self, inp: ConcreteArtifact[S, ArgsT]) -> ConcreteArtifact[T, ArgsT]:
        return ConcreteArtifact(data=self.step(inp.data), args=inp.args)


@dataclasses.dataclass(frozen=True)
class ArgsOnlyAdapter(
    workflow.ChainableWorkflowMixin,
    workflow.ReplaceEnabledWorkflowMixin,
    workflow.Workflow[ConcreteArtifact[DefT, S], ConcreteArtifact[DefT, T]],
    Generic[DefT, S, T],
):
    step: workflow.Workflow[S, T]

    def __call__(self, inp: ConcreteArtifact[DefT, S]) -> ConcreteArtifact[DefT, T]:
        return ConcreteArtifact(data=inp.data, args=self.step(inp.args))


@dataclasses.dataclass(frozen=True)
class StripArgsAdapter(
    workflow.ChainableWorkflowMixin,
    workflow.ReplaceEnabledWorkflowMixin,
    workflow.Workflow[ConcreteArtifact[S, ArgsT], T],
    Generic[ArgsT, S, T],
):
    step: workflow.Workflow[S, T]

    def __call__(self, inp: ConcreteArtifact[S, ArgsT]) -> T:
        return self.step(inp.data)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from dace.sdfg.analysis.schedule_tree import treenodes as tn


def list_index(collection: list[tn.ScheduleTreeNode], node: tn.ScheduleTreeNode) -> int:
    """
    Get the index of `node` in `collection` with `is` operator. Raises `StopIteration` if not found.

    Comparing with the `is` operator ensures memory comparison. The function `list.index()`
    uses value comparison and might thus yield different results.
    """
    # compare with "is" to get memory comparison. ".index()" uses value comparison
    return next(index for index, element in enumerate(collection) if element is node)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import importlib.util
import pathlib
import sys
from types import ModuleType


def import_from_path(
    module_file: pathlib.Path, *, add_to_sys_modules: bool = True, sys_modules_prefix: str = ""
) -> ModuleType:
    """
    Load dynamically a python module from a given file path.

    Args:
        module_file: The path to the python file to load as a module.
        add_to_sys_modules: Whether to add the loaded module to 'sys.modules'.
        sys_modules_prefix: If 'add_to_sys_modules' is `True`, the prefix to use
            for the module name in 'sys.modules'. If empty, the module name will be
            the file name without extension.

    Returns:
        The loaded module.
    """
    module_name = module_file.name.split(".")[0]

    error_msg = f"Could not load module named {module_name} from {module_file}"
    spec = importlib.util.spec_from_file_location(module_name, module_file)
    if not spec or not spec.loader:
        raise ModuleNotFoundError(error_msg)
    try:
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    except ImportError as ex:
        raise ModuleNotFoundError(error_msg) from ex

    if add_to_sys_modules:
        if not sys_modules_prefix.endswith("."):
            sys_modules_prefix += "."
        sys.modules[sys_modules_prefix + module_name] = module
    elif sys_modules_prefix:
        raise ValueError("Cannot use 'sys_modules_prefix' if 'add_to_sys_modules' is False")

    return module



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause
"""An interface between DaCe's MapFusion and the one of GT4Py."""

# Copyright 2019-2021 ETH Zurich and the DaCe authors. All rights reserved.

from typing import Any, Callable, Optional, TypeAlias, Union

import dace
from dace import nodes as dace_nodes, properties as dace_properties
from dace.transformation import dataflow as dace_dftrans


VerticalMapFusionCallback: TypeAlias = Callable[
    ["MapFusionVertical", dace_nodes.MapExit, dace_nodes.MapEntry, dace.SDFGState, dace.SDFG], bool
]
"""Callback used to influence the behaviour of `MapFusionVertical`.

The callback is called by `can_be_applied()` before any other check, thus it is not sure
if the two Maps can be fused at all. If `True` is returned the function will call the
`can_be_applied()` function of the actual Map fusion transformation. If `False` is
returned then `can_be_applied()` will immediately return `False`, i.e. the Maps will
not be checked any further.
It is very similar to the `HorizontalMapFusionCallback` but has different arguments.

It has the following arguments:
- The transformation object itself.
- The MaxExit node of the first Map.
- The MapEntry node of the second Map.
- The SDFGState in which the nodes where found.
- The SDFG itself.
"""

HorizontalMapFusionCallback: TypeAlias = Callable[
    ["MapFusionHorizontal", dace_nodes.MapEntry, dace_nodes.MapEntry, dace.SDFGState, dace.SDFG],
    bool,
]
"""Callback used to influence the behaviour of `MapFusionHorizontal`.

The callback is called by `can_be_applied()` before any other check, thus it is not sure
if the two Maps can be fused at all. If `True` is returned the function will call the
`can_be_applied()` function of the actual Map fusion transformation. If `False` is
returned then `can_be_applied()` will immediately return `False`, i.e. the Maps will
not be checked any further.
It is very similar to the `VerticalMapFusionCallback` but has different arguments.

It has the following arguments:
- The transformation object itself.
- The MaxEntry node of the first Map.
- The MapEntry node of the second Map.
- The SDFGState in which the nodes where found.
- The SDFG itself.
"""


@dace_properties.make_properties
class MapFusionVertical(dace_dftrans.MapFusionVertical):
    """GT4Py's Map fusion transformation for vertical Maps.

    Essentially the same as DaCe's `MapFusionVertical`, but with the following differences:
    - `strict_dataflow` is by default `False`, whereas in DaCe it is `True` for compatibility
        reasons.
    - It accepts the additional `check_fusion_callback`, which is called before the
        `can_be_applied()` function of the DaCe version is called. If it returns `False` then the
        fusion will be rejected.
    """

    _check_fusion_callback: Optional[VerticalMapFusionCallback]

    def __init__(
        self,
        strict_dataflow: bool = False,
        check_fusion_callback: Optional[VerticalMapFusionCallback] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            strict_dataflow=strict_dataflow,
            **kwargs,
        )
        self._check_fusion_callback = None
        if check_fusion_callback is not None:
            self._check_fusion_callback = check_fusion_callback

    def can_be_applied(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        if self._check_fusion_callback is not None:
            if not self._check_fusion_callback(
                self, self.first_map_exit, self.second_map_entry, graph, sdfg
            ):
                return False
        return super().can_be_applied(graph, expr_index, sdfg, permissive)


@dace_properties.make_properties
class MapFusionHorizontal(dace_dftrans.MapFusionHorizontal):
    """GT4Py's Map fusion transformation for horizontal Maps.

    Essentially the same as DaCe's `MapFusionHorizontal`, but with the following difference:
    - It accepts the additional `check_fusion_callback`, which is called before the
        `can_be_applied()` function of the DaCe version is called. If it returns `False` then the
        fusion will be rejected.
    """

    _check_fusion_callback: Optional[HorizontalMapFusionCallback]

    def __init__(
        self,
        check_fusion_callback: Optional[HorizontalMapFusionCallback] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            **kwargs,
        )
        self._check_fusion_callback = None
        if check_fusion_callback is not None:
            self._check_fusion_callback = check_fusion_callback

    def can_be_applied(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        if self._check_fusion_callback is not None:
            if not self._check_fusion_callback(
                self, self.first_parallel_map_entry, self.second_parallel_map_entry, graph, sdfg
            ):
                return False
        return super().can_be_applied(graph, expr_index, sdfg, permissive)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import Final, Iterator, Optional, Sequence, TypeVar

from gt4py.eve import (
    datamodels as eve_datamodels,
    extended_typing as xtyping,
    type_definitions as eve_types,
)
from gt4py.next import common


class TypeSpec(eve_datamodels.DataModel, kw_only=False, frozen=True): ...  # type: ignore[call-arg]


class DataType(TypeSpec):
    """
    A base type for all types that represent data storage.

    Derive floating point, integral or field types from this class.
    """


class CallableType(TypeSpec):
    """
    A base type for all types are callable.

    Derive other callable types, such as FunctionType or FieldOperatorType from
    this class.
    """


class DeferredType(TypeSpec):
    """Dummy used to represent a type not yet inferred."""

    constraint: Optional[type[TypeSpec] | tuple[type[TypeSpec], ...]]


class VoidType(TypeSpec):
    """
    Return type of a function without return values.

    Note: only useful for stateful dialects.
    """


class DimensionType(TypeSpec):
    dim: common.Dimension

    def __str__(self) -> str:
        return str(self.dim)


class IndexType(TypeSpec):
    """
    Represents the type of an index into a dimension.
    """

    dim: common.Dimension

    def __str__(self) -> str:
        return f"Index[{self.dim}]"


class OffsetType(TypeSpec):
    # TODO(havogt): replace by ConnectivityType
    source: common.Dimension
    target: tuple[common.Dimension] | tuple[common.Dimension, common.Dimension]

    def __str__(self) -> str:
        return f"Offset[{self.source}, {self.target}]"


class ScalarKind(eve_types.IntEnum):
    BOOL = 1
    INT8 = 2
    UINT8 = 3
    INT16 = 4
    UINT16 = 5
    INT32 = 6
    UINT32 = 7
    INT64 = 8
    UINT64 = 9
    FLOAT32 = 10
    FLOAT64 = 11
    STRING = 12


class ScalarType(DataType):
    kind: ScalarKind
    shape: Optional[list[int]] = None

    def __str__(self) -> str:
        kind_str = self.kind.name.lower()
        if self.shape is None:
            return kind_str
        return f"{kind_str}{self.shape}"


class ListType(DataType):
    """Represents a neighbor list in the ITIR representation.

    Note:
      - not used in the frontend. The concept is represented as Field with local Dimension.
      - `None` is used to describe lists originating from `make_const_list`.
    """

    element_type: DataType
    offset_type: common.Dimension | None


class FieldType(DataType, CallableType):
    dims: list[common.Dimension]
    dtype: ScalarType | ListType

    def __str__(self) -> str:
        dims = "..." if self.dims is Ellipsis else f"[{', '.join(dim.value for dim in self.dims)}]"
        return f"Field[{dims}, {self.dtype}]"

    @eve_datamodels.validator("dims")
    def _dims_validator(
        self, attribute: eve_datamodels.Attribute, dims: list[common.Dimension]
    ) -> None:
        common.check_dims(dims)


class TupleType(DataType):
    # TODO(tehrengruber): Remove `DeferredType` again. This was erroneously
    #  introduced before we checked the annotations at runtime. All attributes of
    #  a type that are types themselves must be concrete.
    types: list[DataType | DimensionType | DeferredType]

    def __str__(self) -> str:
        return f"tuple[{', '.join(map(str, self.types))}]"

    def __iter__(self) -> Iterator[DataType | DimensionType | DeferredType]:
        yield from self.types

    def __len__(self) -> int:
        return len(self.types)


class AnyPythonType:
    """Marker type representing any Python type which cannot be used for instantiation.

    This is used as a workaround for missing generic support in the case of passing
    a named collection of fields to a scan where it becomes a named collection of scalars.
    """

    def __init__(self) -> None:
        raise AssertionError("Internal Error: The 'AnyPythonType' should never be instantiated.")


#: The 'ANY_PYTHON_TYPE_NAME' can be used in 'NamedCollectionType.original_python_type' to indicate
#: that any original python type is acceptable that is structurally compatible.
#: This is used as a workaround for missing generic support in the case of passing
#: a named collection of fields to a scan where it becomes a named collection of scalars.
#: Note: 'Any' cannot be instantiated and therefore should only be used for type-checking,
#: but not in places where the original python type is actually needed,
#: e.g. `make_named_collection_constructor_from_type_spec`.
ANY_PYTHON_TYPE_NAME: Final[str] = "typing:Any"


class NamedCollectionType(DataType):
    types: list[DataType | DimensionType | DeferredType]
    keys: list[str]
    #: The original python type. It should be only used in the boundaries between
    #: Python and GT4Py DSL, that is, `type translation` and in constructor/extractor
    #: steps for custom containers.
    #: It uses the "entry-point"-like format required by `pkgutil.resolve_name()`:
    #:   '__module__:__qualname__'
    original_python_type: (
        str  # Format: '__module__:__qualname__' (as required by `pkgutil.resolve_name()`)
    )

    def __getattr__(self, name: str) -> DataType | DimensionType | DeferredType:
        keys = object.__getattribute__(self, "keys")
        if name in keys:
            return self.types[keys.index(name)]
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def __str__(self) -> str:
        return f"NamedTuple{{{', '.join(f'{k}: {v}' for k, v in zip(self.keys, self.types))}}}"

    def __iter__(self) -> Iterator[DataType | DimensionType | DeferredType]:
        # Note: Unlike `Mapping`s, we iterate the values (not the keys) by default.
        yield from self.types

    def __len__(self) -> int:
        return len(self.types)


CollectionTypeSpecT = TypeVar("CollectionTypeSpecT", TupleType, NamedCollectionType)
CollectionTypeSpec = TupleType | NamedCollectionType
COLLECTION_TYPE_SPECS: Final[tuple[type[CollectionTypeSpec], ...]] = xtyping.get_args(
    CollectionTypeSpec
)


class FunctionType(CallableType):
    pos_only_args: Sequence[TypeSpec]
    pos_or_kw_args: dict[str, TypeSpec]
    kw_only_args: dict[str, TypeSpec]
    returns: TypeSpec

    def __str__(self) -> str:
        arg_strs = [str(arg) for arg in self.pos_only_args]
        kwarg_strs = [f"{key}: {value}" for key, value in self.pos_or_kw_args.items()]
        args_str = ", ".join((*arg_strs, *kwarg_strs))
        return f"({args_str}) -> {self.returns}"


class ConstructorType(CallableType):
    definition: FunctionType

    @property
    def constructed_type(self) -> TypeSpec:
        return self.definition.returns


class DomainType(DataType):
    dims: list[common.Dimension]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Sequence

from gt4py.next.otf import cpp_utils
from gt4py.next.otf.binding import interface
from gt4py.next.type_system import type_info as ti, type_specifications as ts


def render_function_declaration(function: interface.Function, body: str) -> str:
    template_params: list[str] = []
    rendered_params: list[str] = []
    for index, param in enumerate(function.parameters):
        if isinstance(param.type_, ts.ScalarType):
            rendered_params.append(f"{cpp_utils.pytype_to_cpptype(param.type_)} {param.name}")
        elif ti.is_type_or_tuple_of_type(param.type_, (ts.FieldType, ts.ScalarType)):
            template_param = f"ArgT{index}"
            template_params.append(f"class {template_param}")
            rendered_params.append(f"{template_param}&& {param.name}")
        else:
            raise ValueError(f"Type '{param.type_}' is not supported in C++ interfaces.")

    rendered_decl = f"""decltype(auto) {function.name}({", ".join(rendered_params)}) {{
        {body}
    }}"""
    if template_params:
        return f"""
        template <{", ".join(template_params)}>
        {rendered_decl}
        """.strip()
    return rendered_decl


def render_function_call(function: interface.Function, args: Sequence[str]) -> str:
    return f"{function.name}({', '.join(args)})"



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Fast access to the auto optimization on DaCe."""

import enum
import warnings
from typing import Any, Callable, Optional, Sequence, TypeAlias, Union

import dace
from dace import data as dace_data
from dace.sdfg import nodes as dace_nodes, propagation as dace_propagation, utils as dace_sdutils
from dace.transformation import dataflow as dace_dataflow
from dace.transformation.auto import auto_optimize as dace_aoptimize
from dace.transformation.passes import analysis as dace_analysis

from gt4py.next import common as gtx_common, config as gtx_config
from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


class GT4PyAutoOptHook(enum.Enum):
    """Allows to hook into different stages of the auto optimizer.

    The hook system allows to inject certain additional behaviour at various places
    during the execution of `gt_auto_optimize()`. See the list below for more specific
    instructions. Most of the hooks are expected to conform to the signature specified
    by `GT4PyAutoOptHookStage`. However, some hooks require a different signature.

    The supported values are:
    - TopLevelDataFlowPre: Called before the top level dataflow is optimized.
    - TopLevelDataFlowStep: Called during the top level optimization stage, which might be
        called multiple times.
    - TopLevelDataFlowMapFusionVerticalCallBack: If provided this function is passed
        as `check_fusion_callback` argument to `MapFusionVertical` during the top level
        dataflow optimization stage. This hook is also forwarded to the
        `gt_vertical_map_split_fusion()` function, which is called with
        `fuse_map_fragments=False` but `run_map_fusion=True`. Note that this hook has
        to meet the requirements of `VerticalMapFusionCallback`, see the `map_fusion`
        module for more information.
    - TopLevelDataFlowMapFusionHorizontalCallBack: If provided this function is passed
        as `check_fusion_callback` argument to `MapFusionHorizontal` during the top
        level dataflow optimization. This hook is also forwarded to the
        `gt_horizontal_map_split_fusion()` function, which is called with
        `fuse_map_fragments=False` but `run_map_fusion=True`. Note that this hook has
        to meet the requirements of `HorizontalMapFusionCallback`, see the `map_fusion`
        module for more information.
    - TopLevelDataFlowMapPromotionCallBack: If provided then pass this function as
        `promotion_callback` argument to the `MapPromoter` during top level dataflow
        optimization. If not provided then `gt4py_default_auto_optimizer_map_promotion_checker()`
        is used. Note that this callback has to meet the requirements of
        `MapPromotionCallBack`.
    - `TopLevelDataFlowVerticalSplitCallBack`: If provided it is passed as `check_split_callback`
        to `gt_vertical_map_split_fusion()` and allows to control how a vertical Map
        split is done. See `VerticalMapSplitCallback` for more details.
        Note that the function is called with `fuse_map_fragments=False` but
        `run_map_fusion=True`.
    - `TopLevelDataFlowHorizontalSplitCallBack`: If provided it is passed as `check_split_callback`
        to `gt_horizontal_map_split_fusion()` and allows to control how a horizontal
        Map split is done. See `HorizontalMapSplitCallback` for more details.
        Note that the function is called with `fuse_map_fragments=False` but
        `run_map_fusion=True`.
    - TopLevelDataFlowPost: Called after the top level dataflow has been optimized.
    - AfterToGPU: This hook is called after the GPU transformation has been called,
        if the SDFG should run on GPU.

    Note:
        The `TopLevelDataFlowMapFusionVerticalCallBack` and `TopLevelDataFlowMapFusionHorizontalCallBack`
        hooks are not passed to the splitting transformations, this is a limitation of
        the toolchain and should be fixed soon. If they are used it is advised to
        pass `disable_splitting` to `gt_auto_optimize()` and call the splitting
        transformations directly if desired.

    Todo:
        - Implement a hook for the splitting transformation.
    """

    TopLevelDataFlowPre = enum.auto()
    TopLevelDataFlowMapFusionVerticalCallBack = enum.auto()
    TopLevelDataFlowMapFusionHorizontalCallBack = enum.auto()
    TopLevelDataFlowMapPromotionCallBack = enum.auto()
    TopLevelDataFlowVerticalSplitCallBack = enum.auto()
    TopLevelDataFlowHorizontalSplitCallBack = enum.auto()
    TopLevelDataFlowStep = enum.auto()
    TopLevelDataFlowPost = enum.auto()
    AfterToGPU = enum.auto()


GT4PyAutoOptHookStage: TypeAlias = Callable[[dace.SDFG], None]
"""Signature of a regular "stage hook" of the optimizer.

The function gets the SDFG as its single argument and processes it in-place.
Most of the hooks `gt_auto_optimize()` provides are expected to be of that type.
"""


GT4PyAutoOptHookFun: TypeAlias = Union[
    GT4PyAutoOptHookStage,
    "gtx_transformations.HorizontalMapFusionCallback",
    "gtx_transformations.VerticalMapFusionCallback",
    "gtx_transformations.MapPromotionCallBack",
    "gtx_transformations.VerticalMapSplitCallback",
    "gtx_transformations.HorizontalMapSplitCallback",
]


def gt_auto_optimize(
    sdfg: dace.SDFG,
    gpu: bool,
    unit_strides_kind: Optional[gtx_common.DimensionKind] = None,
    make_persistent: bool = False,
    gpu_block_size: Optional[Sequence[int | str] | str] = (32, 8, 1),
    gpu_block_size_1d: Optional[Sequence[int | str] | str] = (64, 1, 1),
    gpu_block_size_2d: Optional[Sequence[int | str] | str] = None,
    gpu_block_size_3d: Optional[Sequence[int | str] | str] = None,
    gpu_maxnreg: Optional[int] = None,
    blocking_dim: Optional[gtx_common.Dimension] = None,
    blocking_size: int = 10,
    blocking_only_if_independent_nodes: bool = True,
    scan_loop_unrolling: bool = False,
    scan_loop_unrolling_factor: int = 0,
    disable_splitting: bool = False,
    reuse_transients: bool = False,
    gpu_launch_bounds: Optional[int | str] = None,
    gpu_launch_factor: Optional[int] = None,
    gpu_memory_pool: bool = True,
    constant_symbols: Optional[dict[str, Any]] = None,
    assume_pointwise: bool = True,
    optimization_hooks: Optional[dict[GT4PyAutoOptHook, GT4PyAutoOptHookFun]] = None,
    demote_fields: Optional[list[str]] = None,
    fuse_tasklets: bool = False,
    validate: bool = True,
    validate_all: bool = False,
    **kwargs: Any,
) -> dace.SDFG:
    """Performs GT4Py specific optimizations on the SDFG in place.

    This function expects that the input SDFG follows the principles that are
    outlined in [ADR0018](https://github.com/GridTools/gt4py/tree/main/docs/development/ADRs/0018-Canonical_SDFG_in_GT4Py_Transformations.md).
    It is important to notice, that if `reuse_transients` is active then the
    optimized SDFG no longer conforms to these rules.

    The auto optimization works in different phases, that focuses each on
    different aspects of the SDFG. The initial SDFG is assumed to have a
    very large number of rather simple Maps.

    Note, because of how `gt_auto_optimizer()` works it is not save to call
    it twice on the same SDFG.

    A user should not rely on a particular order in which the SDFG is processed.
    However, in broad terms the function performs roughly the following steps:
    - An initial simplification is performed. This also includes the substitution
        of symbols whose value is known, see `constant_symbols`.
    - The top level Maps are optimized, the goal of this step is to increase
        the computational intensity. This is mostly, but not limited to fusing
        then together.
    - The interior dataflow of the top level Maps are optimized.
    - The Maps are configured. This involves setting their iteration order, such
        that the unit stride dimension is associated to the `x` block (GPU) or
        inner most loop-nest (CPU). But it also involves moving everything to GPU,
        if needed and setting the block size.
    - The transients are processed, this involves steps such as determine their
        strides or if they are persistently allocated.

    Args:
        sdfg: The SDFG that should be optimized in place.
        gpu: Optimize for GPU or CPU.
        unit_strides_kind: All dimensions of this kind are considered to have unit
            strides, see `gt_set_iteration_order()` for more.
        make_persistent: Turn all transients to persistent lifetime, thus they are
            allocated over the whole lifetime of the program, even if the kernel exits.
            Thus the SDFG can not be called by different threads.
        gpu_block_size: This is used as default thread block size for GPU Maps. See
            also the `gpu_block_size_*d` arguments
        gpu_block_size_{1, 2, 3}d: Allows to specify the GPU thread block size for
            1, 2 and 3 dimension Maps individually. See the `gpu_block_size_spec`
            argument of `gt_gpu_transformation()` for more.
        blocking_dim: On which dimension blocking should be applied.
        blocking_size: How many elements each block should process.
        blocking_only_if_independent_nodes: If `True`, the default, only apply loop
            blocking if there are independent nodes in the Map, see the
            `require_independent_nodes` option of the `LoopBlocking` transformation.
        scan_loop_unrolling: Whether to unroll scan loops.
        scan_loop_unrolling_factor: The unroll factor to use when unrolling scan loops.
        disable_splitting: Disable the splitting transformations.
        reuse_transients: Run the `TransientReuse` transformation, might reduce memory footprint.
        gpu_launch_bounds: Use this value as `__launch_bounds__` for _all_ GPU Maps.
        gpu_launch_factor: Use the number of threads times this value as `__launch_bounds__`
            for _all_ GPU Maps.
        gpu_memory_pool: Enable CUDA memory pool in gpu codegen.
        constant_symbols: Symbols listed in this `dict` will be replaced by the
            respective value inside the SDFG. This might increase performance.
        assume_pointwise: Assume that the SDFG has no risk for race condition in
            global data access. See the `GT4PyMapBufferElimination` transformation for more.
        optimization_hooks: A `dict` containing the hooks that should be called,
            see `GT4PyAutoOptHook` for more information.
        demote_fields: Consider these fields as transients for the purpose of optimization.
            Use at your own risk. See Notes for all implications.
        fuse_tasklets: Reduces the number of Tasklets by fusing them.
        validate: Perform validation during the steps.
        validate_all: Perform extensive validation.

    Note:
        - For identifying symbols that can be treated as compile time constants
            `gt_find_constant_arguments()` function can be used.
        - When `unit_strides_kind` is `None` the function assumes that on GPU the
            horizontal dimension has unit strides, while on CPU the vertical dimension
            has unit strides.
        - In case GPU optimizations are enabled, the function assumes that all
            global fields are already on the GPU.
        - When using the `demote_fields` feature the user has to ensure that the fields
            meet all constraints of a transient, see ADR18 for more. Furthermore, there
            is no guarantee whether the demoted field remains part of the SDFG's call
            signature or not.

    Todo:
        - Update the description. The Phases are nice, but they have lost their
            link to reality a little bit.
        - Improve the determination of the strides and iteration order of the
            transients.
        - Set padding of transients, i.e. alignment, the DaCe datadescriptor
            can do that.
        - Handle nested SDFGs better.
        - Specify arguments to set the size of GPU thread blocks depending on the
            dimensions. I.e. be able to use a different size for 1D than 2D Maps.
        - Implement some model to further guide to determine what we want to fuse.
            Something along the line "Fuse if operational intensity goes up, but
            not if we have too much internal space (register pressure).
    """
    device = dace.DeviceType.GPU if gpu else dace.DeviceType.CPU
    optimization_hooks = optimization_hooks or {}

    with dace.config.temporary_config():
        # Do not print the transformation progress, unless enabled through env variable.
        dace.Config.set(
            "progress", value=gtx_config.env_flag_to_bool("DACE_progress", default=gtx_config.DEBUG)
        )

        # Do not store which transformations were applied inside the SDFG.
        dace.Config.set("store_history", value=False)

        # If there is an exception during `can_be_applied()` propagate it instead
        #  of ignoring it.
        dace.Config.set("optimizer", "match_exception", value=True)

        # Do not assume that symbols are positive. The only reason for this is
        #  that the origin can potentially be negative.
        #  See also [issue#2095](https://github.com/spcl/dace/issues/2095)
        dace.Config.set("optimizer", "symbolic_positive", value=False)

        # Initial Cleanup
        # NOTE: The initial simplification stage must be synchronized with the one that
        #   `gt_substitute_compiletime_symbols()` performs!
        gtx_transformations.gt_simplify(
            sdfg=sdfg,
            validate=False,
            skip=gtx_transformations.constants._GT_AUTO_OPT_INITIAL_STEP_SIMPLIFY_SKIP_LIST,
            validate_all=validate_all,
        )

        if constant_symbols:
            gtx_transformations.gt_substitute_compiletime_symbols(
                sdfg=sdfg,
                repl=constant_symbols,
                simplify=True,  # Simplify again after.
                skip=gtx_transformations.constants._GT_AUTO_OPT_INITIAL_STEP_SIMPLIFY_SKIP_LIST,
                simplify_at_entry=False,
                validate=False,
                validate_all=validate_all,
            )

        # Demote the fields.
        #  Actually they should probably be at the very start of this function, however,
        #  they have to be after constant substitution in case the descriptor is modified.
        original_demoted_descriptors: dict[str, dace_data.Data] = {}
        if demote_fields is not None:
            for field_to_demote in demote_fields:
                if field_to_demote not in sdfg.arrays:
                    warnings.warn(
                        f"Requested the demotion of field '{field_to_demote}' but the field is unknown.",
                        stacklevel=0,
                    )
                    continue
                field_desc = sdfg.arrays[field_to_demote]

                if field_desc.transient:
                    warnings.warn(
                        f"Requested the demotion of field '{field_to_demote}' but the field is a transient.",
                        stacklevel=0,
                    )
                    continue

                # Demote the field but keep a copy of it such that we can restore it later.
                original_demoted_descriptors[field_to_demote] = field_desc.clone()
                field_desc.transient = True

            if len(original_demoted_descriptors) != 0:
                gtx_transformations.gt_simplify(
                    sdfg=sdfg,
                    validate=False,
                    skip=gtx_transformations.constants._GT_AUTO_OPT_INITIAL_STEP_SIMPLIFY_SKIP_LIST,
                    validate_all=validate_all,
                )

        gtx_transformations.gt_reduce_distributed_buffering(
            sdfg, validate=False, validate_all=validate_all
        )

        # Process top level Maps
        sdfg = _gt_auto_process_top_level_maps(
            sdfg=sdfg,
            assume_pointwise=assume_pointwise,
            disable_splitting=disable_splitting,
            optimization_hooks=optimization_hooks,
            validate_all=validate_all,
        )

        # We now ensure that point wise computations are properly double buffered,
        #  this ensures that rule 3 of ADR-18 is maintained.
        # TODO(phimuell): Figuring out if it is important to do it before the inner
        #   Map optimization. I think it is, especially when we apply `LoopBlocking`.
        gtx_transformations.gt_create_local_double_buffering(sdfg)

        # Optimize the interior of the Maps:
        sdfg = _gt_auto_process_dataflow_inside_maps(
            sdfg=sdfg,
            blocking_dim=blocking_dim,
            blocking_size=blocking_size,
            blocking_only_if_independent_nodes=blocking_only_if_independent_nodes,
            scan_loop_unrolling=scan_loop_unrolling,
            scan_loop_unrolling_factor=scan_loop_unrolling_factor,
            fuse_tasklets=fuse_tasklets,
            validate_all=validate_all,
        )

        # Configure the Maps:
        #  Will also perform the GPU transformation.
        # TODO(phimuell): Maybe switch it with the inside map optimization.
        gpu_block_size_spec: dict[str, Sequence[int | str] | str] = {}
        if gpu_block_size_1d is not None:
            gpu_block_size_spec["block_size_1d"] = gpu_block_size_1d
        if gpu_block_size_2d is not None:
            gpu_block_size_spec["block_size_2d"] = gpu_block_size_2d
        if gpu_block_size_3d is not None:
            gpu_block_size_spec["block_size_3d"] = gpu_block_size_3d

        # We now turn the demoted fields back into globals if possible, for ABI compatibility.
        #  We have to do it here, because we need to know later if something is a transient
        #  or not.
        for demoted_field, original_field_desc in original_demoted_descriptors.items():
            if demoted_field not in sdfg.arrays:
                # The demoted field is not inside the SDFG anymore, just insert it.
                sdfg.add_datadesc(demoted_field, original_field_desc)

            elif original_field_desc.is_equivalent(sdfg.arrays[demoted_field]) and all(
                (ostride == cstride) == True  # noqa: E712 [true-false-comparison]  # SymPy comparison
                for ostride, cstride in zip(
                    sdfg.arrays[demoted_field].strides, original_field_desc.strides
                )
            ):
                # The demoted field is still inside the SDFG, we only add it back if it
                #  still has the same layout as before.
                # NOTE: Technically we could ignore the strides, the problem are nested
                #   SDFG, where they might be mapped into.
                sdfg.arrays[demoted_field].transient = False

            else:
                warnings.warn(
                    f"Could not restore the demoted field '{demoted_field}' back to a global.",
                    stacklevel=0,
                )

        sdfg = _gt_auto_configure_maps_and_strides(
            sdfg=sdfg,
            gpu=gpu,
            unit_strides_kind=unit_strides_kind,
            gpu_block_size=gpu_block_size,
            gpu_launch_factor=gpu_launch_factor,
            gpu_launch_bounds=gpu_launch_bounds,
            gpu_maxnreg=gpu_maxnreg,
            optimization_hooks=optimization_hooks,
            gpu_block_size_spec=gpu_block_size_spec if gpu_block_size_spec else None,
            validate_all=validate_all,
        )

        # Transients
        sdfg = _gt_auto_post_processing(
            sdfg=sdfg,
            gpu=gpu,
            make_persistent=make_persistent,
            # TODO(phimuell): In general `TransientReuse` is a good idea, but the
            #   current implementation also reuses transients scalars inside Map
            #   scopes, which I do not like. Thus we should fix the transformation
            #   to avoid that.
            reuse_transients=reuse_transients,
            gpu_memory_pool=gpu_memory_pool,
            validate_all=validate_all,
        )

        # Canonicalize the SDFG. This ensures that the code generator will see SDFGs
        #  that conform to the historical expected version.
        # NOTE: Currently needed to avoid an error in `AddThreadBlockMap`.
        dace.sdfg.utils.canonicalize_memlet_trees(sdfg)

        # Set the implementation of the library nodes.
        dace_aoptimize.set_fast_implementations(sdfg, device)

        if validate:
            sdfg.validate()

        return sdfg


def _gt_auto_process_top_level_maps(
    sdfg: dace.SDFG,
    assume_pointwise: bool,
    optimization_hooks: dict[GT4PyAutoOptHook, GT4PyAutoOptHookFun],
    disable_splitting: bool,
    validate_all: bool,
) -> dace.SDFG:
    """Optimize the Maps at the top level of the SDFG inplace.

    In broad terms the function tries to reduce the number of of kernels/maps
    by fusing/inlining them, however, the function will also perform some other
    transformations, with the goal to promote high computational intensity and
    occupancy (on GPU).

    It is important to note that the fusion will only affect top level maps, i.e.
    everything whose scope is `None`. Some parts of the optimization process are
    iteratively.

    For a full description of the argument see the `gt_auto_optimize()` function.

    Note:
        The function assumes that `gt_simplify()` has been called on the SDFG
        before it is passed to this function.
    """

    # NOTE: Inside this function we have to disable the consolidation of edges.
    #   This is because it might block the application of `SpliAccessNode`. As
    #   an example consider a Map that writes `tmp[:, 80]` and a second Map that
    #   writes `tmp[:, 0]`, if these two Maps are now horizontally fussed and
    #   edge consolidation is on, then the resulting Map would "write", at least
    #   according to the subset, after Memlet propagation, into `tmp[:, 0:80]`.
    #   For that reason we block edge consolidation inside this function.
    #   However, we allow allow the consolidation, in MapFusion if this does
    #   not lead to an extension. This is because it causes some issues with
    #   MapFusion.
    # TODO(phimuell): Find a better way as blocking edge consolidation might
    #   limit what MapFusion can do.
    # TODO(phimuell): Maybe disable edge consolidation by default?

    # Compute the SDFG hash to see if something has changed.
    #  We use the hash instead of the return values of the transformation, because
    #  computing the hash invalidates some caches that are not properly updated in Dace.
    # TODO(phimuell): Remove this hack as soon as DaCe is fixed.
    # TODO(phimuell): Maybe switch to `reset_cfg_list()`?
    sdfg_hash = sdfg.hash_sdfg()

    if GT4PyAutoOptHook.TopLevelDataFlowPre in optimization_hooks:
        optimization_hooks[GT4PyAutoOptHook.TopLevelDataFlowPre](sdfg)  # type: ignore[call-arg]

    while True:
        # First we do scan the entire SDFG to figure out which data is only
        #  used once and can be deleted.
        find_single_use_data = dace_analysis.FindSingleUseData()
        single_use_data = find_single_use_data.apply_pass(sdfg, None)

        # TODO(phimuell): Use a cost measurement to decide if fusion should be done.

        vertical_map_fusion = gtx_transformations.MapFusionVertical(
            only_toplevel_maps=True,
            consolidate_edges_only_if_not_extending=True,
            check_fusion_callback=optimization_hooks.get(  # type: ignore[arg-type]
                GT4PyAutoOptHook.TopLevelDataFlowMapFusionVerticalCallBack, None
            ),
        )
        # TODO(phimuell): Remove that hack once [issue#1911](https://github.com/spcl/dace/issues/1911)
        #   has been solved.
        vertical_map_fusion._single_use_data = single_use_data

        sdfg.apply_transformations_repeated(
            vertical_map_fusion,
            validate=False,
            validate_all=validate_all,
        )

        # Try to remove some unnecessary copies.
        sdfg.apply_transformations_repeated(
            gtx_transformations.DoubleWriteRemover(
                single_use_data=single_use_data,
            ),
            validate=False,
            validate_all=validate_all,
        )

        # NOTE: There is a Memlet caching issue at work here, see DaCe issue 1703 and
        #   1708. Without clearing the cache, which is done through a side effect of
        #   `to_json()`, running `propagate_memlets_sdfg()` would lead to an invalid
        #   SDFG.
        sdfg.to_json(hash=False)

        # Promote Maps. This will remove transients between 1D and 2D Maps, at the
        #  cost of more data loads from memory. Empirical observations have shown
        #  that this is beneficial; especially for Nabla4-type kernel in conjunction
        #  with `LoopBlocking`.
        # NOTE: We have to promote before we do horizontal Map fusion.
        sdfg.apply_transformations_repeated(
            gtx_transformations.MapPromoter(
                only_toplevel_maps=True,
                promote_vertical=True,
                promote_horizontal=True,
                promote_local=False,
                promotion_callback=optimization_hooks.get(
                    GT4PyAutoOptHook.TopLevelDataFlowMapPromotionCallBack,
                    gt4py_default_auto_optimizer_map_promotion_checker,
                ),
                single_use_data=single_use_data,
            ),
            validate=False,
            validate_all=validate_all,
        )

        # Now run horizontal Map fusion, also run vertical Map fusion, because it
        #  might have become possible.
        # TODO(phimuell): Think of moving horizontal Map fusion after the splitting.
        horizontal_map_fusion = gtx_transformations.MapFusionHorizontal(
            only_toplevel_maps=True,
            consolidate_edges_only_if_not_extending=True,
            check_fusion_callback=optimization_hooks.get(  # type: ignore[arg-type]
                GT4PyAutoOptHook.TopLevelDataFlowMapFusionHorizontalCallBack, None
            ),
            # NOTE: Setting this argument to `False` has the side effect that Maps
            #   without any connection to each other are fused together and by this
            #   creating dependencies between otherwise unrelated Maps. However,
            #   also has the side effect of integrating Maps generated in
            #   `broadcast+concat_where` situations into other Maps. We should handle
            #   these cases through the splitting transformations.
            # TODO(phimuell): Update the other transformations such that we can
            #   set the falg to `True`.
            only_if_common_ancestor=False,
        )
        sdfg.apply_transformations_repeated(
            [horizontal_map_fusion, vertical_map_fusion],
            validate=False,
            validate_all=validate_all,
        )

        # Now do some cleanup task, that may enable further fusion opportunities.
        #  Note for performance reasons simplify is deferred.
        gtx_transformations.gt_reduce_distributed_buffering(
            sdfg, validate=False, validate_all=validate_all
        )

        if not disable_splitting:
            # TODO(phimuell): Implement a data cleaner.
            dace_sdutils.canonicalize_memlet_trees(sdfg)
            dace_propagation.propagate_memlets_sdfg(sdfg)
            sdfg.apply_transformations_repeated(
                [
                    gtx_transformations.MapSplitter(
                        single_use_data=single_use_data,
                        remove_dead_dataflow=True,
                    ),
                    gtx_transformations.SplitConsumerMemlet(single_use_data=single_use_data),
                ],
                validate=False,
                validate_all=validate_all,
            )
            # TODO(phimuell): Find out how to skip the propagation and integrating it
            #   into the split transformation.
            dace_sdutils.canonicalize_memlet_trees(sdfg)
            dace_propagation.propagate_memlets_sdfg(sdfg)

            # Split the top level AccessNodes.
            # NOTE: This function will also update `single_use_data`.
            gtx_transformations.gt_split_access_nodes(
                sdfg=sdfg,
                validate=False,
                validate_all=validate_all,
                single_use_data=single_use_data,
            )

            # Perform buffer elimination.
            # TODO(phimuell): Implement a faster matching.
            sdfg.apply_transformations_once_everywhere(
                gtx_transformations.GT4PyMapBufferElimination(
                    assume_pointwise=assume_pointwise,
                ),
                validate=False,
                validate_all=validate_all,
            )

            # Call vertical and horizontal map fusion to fuse together maps on partially
            #  overlapping range. This is an iterative process that splits the maps to
            #  expose overlapping range and applies serial/parallel map fusion.
            gtx_transformations.gt_vertical_map_split_fusion(
                sdfg=sdfg,
                run_simplify=False,
                run_map_fusion=True,
                fuse_map_fragments=False,  # To avoid uncontrolled Map fusing.
                skip=gtx_transformations.constants._GT_AUTO_OPT_TOP_LEVEL_STAGE_SIMPLIFY_SKIP_LIST,
                consolidate_edges_only_if_not_extending=True,
                single_use_data=single_use_data,
                check_split_callback=optimization_hooks.get(  # type: ignore[arg-type]
                    GT4PyAutoOptHook.TopLevelDataFlowVerticalSplitCallBack, None
                ),
                check_fusion_callback=optimization_hooks.get(  # type: ignore[arg-type]
                    GT4PyAutoOptHook.TopLevelDataFlowMapFusionVerticalCallBack, None
                ),
                validate=False,
                validate_all=validate_all,
            )
            gtx_transformations.gt_horizontal_map_split_fusion(
                sdfg=sdfg,
                run_simplify=False,
                run_map_fusion=True,
                fuse_map_fragments=False,  # To avoid uncontrolled Map fusing.
                skip=gtx_transformations.constants._GT_AUTO_OPT_TOP_LEVEL_STAGE_SIMPLIFY_SKIP_LIST,
                consolidate_edges_only_if_not_extending=True,
                check_split_callback=optimization_hooks.get(  # type: ignore[arg-type]
                    GT4PyAutoOptHook.TopLevelDataFlowHorizontalSplitCallBack, None
                ),
                check_fusion_callback=optimization_hooks.get(  # type: ignore[arg-type]
                    GT4PyAutoOptHook.TopLevelDataFlowMapFusionHorizontalCallBack, None
                ),
                validate=False,
                validate_all=validate_all,
            )

        # TODO(phimuell): Deprecate this transformation.
        sdfg.apply_transformations_repeated(
            gtx_transformations.GT4PyMapBufferElimination(
                assume_pointwise=assume_pointwise,
            ),
            validate=False,
            validate_all=validate_all,
        )

        # TODO(phimuell): Figuring out if this is the correct location for doing it.
        if GT4PyAutoOptHook.TopLevelDataFlowStep in optimization_hooks:
            optimization_hooks[GT4PyAutoOptHook.TopLevelDataFlowStep](sdfg)  # type: ignore[call-arg]

        # Determine if the SDFG has been modified by comparing the hash.
        old_sdfg_hash, sdfg_hash = sdfg_hash, sdfg.hash_sdfg()
        if old_sdfg_hash == sdfg_hash:
            break

        # The SDFG was modified by the transformations above. The SDFG was
        #  modified. Call Simplify and try again to further optimize.
        gtx_transformations.gt_simplify(
            sdfg,
            validate=False,
            validate_all=validate_all,
            skip=gtx_transformations.constants._GT_AUTO_OPT_TOP_LEVEL_STAGE_SIMPLIFY_SKIP_LIST,
        )

    # Replace `concat_where` nodes
    # TODO(phimuell): Are there better locations for this transformation?
    gtx_transformations.gt_apply_concat_where_replacement_on_sdfg(
        sdfg=sdfg,
        single_use_data=single_use_data,
        validate=False,
        validate_all=validate_all,
    )

    if GT4PyAutoOptHook.TopLevelDataFlowPost in optimization_hooks:
        optimization_hooks[GT4PyAutoOptHook.TopLevelDataFlowPost](sdfg)  # type: ignore[call-arg]

    return sdfg


def _gt_auto_process_dataflow_inside_maps(
    sdfg: dace.SDFG,
    blocking_dim: Optional[gtx_common.Dimension],
    blocking_size: int,
    blocking_only_if_independent_nodes: Optional[bool],
    scan_loop_unrolling: bool,
    scan_loop_unrolling_factor: int,
    fuse_tasklets: bool,
    validate_all: bool,
) -> dace.SDFG:
    """Optimizes the dataflow inside the top level Maps of the SDFG inplace.

    For a description of the arguments see `gt_auto_optimize()`.

    It is important that nested Maps are not fused together. The main reason is that
    in most cases it is not beneficial and might prevent other optimization, as the
    fusion (especially the parallel version) will create dependency of otherwise
    unrelated dataflow. Furthermore, the majority, if not all, of these maps are
    over a constant range, e.g. the number of neighbours, which is known at compile
    time, so the compiler will fully unroll them anyway.
    """

    # Separate Tasklets into dependent and independent parts to promote data
    #  reusability. It is important that this step has to be performed before
    #  `TaskletFusion` is used.
    if blocking_dim is not None:
        sdfg.apply_transformations_once_everywhere(
            gtx_transformations.LoopBlocking(
                blocking_size=blocking_size,
                blocking_parameter=blocking_dim,
                require_independent_nodes=blocking_only_if_independent_nodes,
            ),
            validate=False,
            validate_all=validate_all,
        )

    # Merge Tasklets into bigger ones.
    # NOTE: Empirical observation for Graupel have shown that this leads to an increase
    #   in performance, however, it has to be run before `GT4PyMoveTaskletIntoMap`
    #   (not fully clear why though, probably a compiler artefact) and as well as
    #   `MoveDataflowIntoIfBody` (not fully clear either, it `TaskletFusion` makes
    #   things simpler or prevent it from doing certain, negative, things).
    # TODO(phimuell): Investigate more.
    # TODO(phimuell): Restrict it to Tasklets only inside Maps.
    if fuse_tasklets:
        sdfg.apply_transformations_repeated(
            dace_dataflow.TaskletFusion,
            validate=False,
            validate_all=validate_all,
        )

    # Constants (tasklets are needed to write them into a variable) should not be
    #  arguments to a kernel but be present inside the body.
    sdfg.apply_transformations_once_everywhere(
        gtx_transformations.GT4PyMoveTaskletIntoMap,
        validate=False,
        validate_all=validate_all,
    )

    # TODO(phimuell): figuring out if this is needed?
    gtx_transformations.gt_simplify(
        sdfg,
        skip=gtx_transformations.constants._GT_AUTO_OPT_INNER_DATAFLOW_STAGE_SIMPLIFY_SKIP_LIST,
        validate=False,
        validate_all=validate_all,
    )

    find_single_use_data = dace_analysis.FindSingleUseData()
    single_use_data = find_single_use_data.apply_pass(sdfg, None)

    sdfg.apply_transformations_repeated(
        gtx_transformations.RemoveScalarCopies(
            single_use_data=single_use_data,
        ),
        validate=False,
        validate_all=validate_all,
    )

    # Make sure that this runs before MoveDataflowIntoIfBody because atm it doesn't handle
    # NestedSDFGs inside the ConditionalBlocks it fuses.
    sdfg.apply_transformations_repeated(
        gtx_transformations.FuseHorizontalConditionBlocks(),
        validate=True,
        validate_all=True,
    )

    # Move dataflow into the branches of the `if` such that they are only evaluated
    #  if they are needed. Important to call it repeatedly.
    # TODO(phimuell): It is unclear if `MoveDataflowIntoIfBody` should be called
    #   before or after `LoopBlocking`. In cases where the condition is `False`
    #   most of the times calling it before is better, but if the condition is
    #   `True` then this order is better. Solve that issue.
    # NOTE: The transformation is currently only able to handle dataflow that is
    #   _directly_ enclosed by a Map. Thus the order in which they (multiple blocks
    #   in the same Map) are processed matter. Think of a chain of `if` blocks that
    #   can be perfectly nested. If the last one is handled first, then all other
    #   can not be processed anymore. This means it is important to set
    #   `ignore_upstream_blocks` to `False`, thus the transformation will not apply
    #   if the dataflow that should be relocated into an `if` block contains again
    #   `if` blocks that can be relocated. It would be more efficient to process
    #   them in the right order from the beginning.
    sdfg.apply_transformations_repeated(
        gtx_transformations.MoveDataflowIntoIfBody(
            ignore_upstream_blocks=False,
        ),
        validate=False,
        validate_all=validate_all,
    )

    # TODO(phimuell): figuring out if this is needed?
    gtx_transformations.gt_simplify(
        sdfg,
        skip=gtx_transformations.constants._GT_AUTO_OPT_INNER_DATAFLOW_STAGE_SIMPLIFY_SKIP_LIST,
        validate=False,
        validate_all=validate_all,
    )

    # After some transformation we see in the SDFG that there are pointwise views
    # generated after reduction nodes. These views are unecessary and might produce
    # sub optimal GPU code thus we remove them.
    sdfg.apply_transformations_once_everywhere(
        gtx_transformations.RemovePointwiseViews,
        validate=False,
        validate_all=validate_all,
    )

    if scan_loop_unrolling:
        sdfg.apply_transformations_once_everywhere(
            gtx_transformations.ScanLoopUnrolling(unroll_factor=scan_loop_unrolling_factor),
            validate=False,
            validate_all=validate_all,
        )

    return sdfg


def _gt_auto_configure_maps_and_strides(
    sdfg: dace.SDFG,
    gpu: bool,
    unit_strides_kind: Optional[gtx_common.DimensionKind],
    gpu_block_size: Optional[Sequence[int | str] | str],
    gpu_launch_bounds: Optional[int | str],
    gpu_launch_factor: Optional[int],
    gpu_maxnreg: Optional[int],
    optimization_hooks: dict[GT4PyAutoOptHook, GT4PyAutoOptHookFun],
    gpu_block_size_spec: Optional[dict[str, Sequence[int | str] | str]],
    validate_all: bool,
) -> dace.SDFG:
    """Configure the Maps and the strides of the SDFG inplace.

    Essentially the function will set the properties of the Map accordingly, most
    importantly the iteration order of the loops, to ensure that the unit stride
    is associated to the `x` dimension of a block (GPU) or the inner most loop-nest.
    Furthermore, it will also apply the GPU transformation.
    It will also determine the strides of the transients.

    For a description of the arguments see the `gt_auto_optimize()` function.
    """

    # If `unit_strides_kind` is unknown we will not modify the Map order nor the
    #  strides, except if we are on GPU. The reason for this is that the maximal
    #  number of blocks is different for each dimension. If the largest dimension
    #  is for example associated with the `z` dimension, we would get launch errors
    #  at some point. Thus in that case we pretend that it is horizontal. Which is
    #  a valid assumption for any ICON-like code or if the GT4Py allocator is used.
    # TODO(phimuell): Make this selection more intelligent.
    if unit_strides_kind is None and gpu:
        prefered_direction_kind: Optional[gtx_common.DimensionKind] = (
            gtx_common.DimensionKind.HORIZONTAL
        )
    else:
        prefered_direction_kind = unit_strides_kind

    # We should actually use a `gtx.Dimension` here and not a `gtx.DimensionKind`,
    #  since they are unique. However at this stage, especially after the expansion
    #  of non standard Memlets (which happens in the GPU transformation) associating
    #  Map parameters with GT4Py dimension is very hard to impossible. At this stage
    #  the kind is the most reliable indicator we have.
    # NOTE: This is not the only location where we manipulate the Map order, we also
    #   do it in the GPU transformation, where we have to set the order of the
    #   expanded Memlets.
    if prefered_direction_kind is not None:
        gtx_transformations.gt_set_iteration_order(
            sdfg=sdfg,
            unit_strides_kind=prefered_direction_kind,
            validate=False,
            validate_all=validate_all,
        )

    # NOTE: We have to set the strides of transients before the non-standard Memlets
    #   get expanded, i.e. turned into Maps because no matching `cudaMemcpy*()` call
    #   exists, which requires that the final strides are there. Furthermore, Memlet
    #   expansion has to happen before the GPU block size is set. There are several
    #   possible solutions for that, of which none is really good. The least bad one
    #   is to set the strides of the transients here. The main downside is that we
    #   slightly modify the SDFG in the GPU transformation after we have set the
    #   strides.
    if prefered_direction_kind is not None:
        gtx_transformations.gt_change_strides(sdfg, prefered_direction_kind=prefered_direction_kind)

    if gpu:
        if unit_strides_kind != gtx_common.DimensionKind.HORIZONTAL:
            warnings.warn(
                "The GT4Py DaCe GPU backend assumes that the leading dimension, i.e."
                " where stride is 1, is of kind 'HORIZONTAL', however it was"
                f" '{unit_strides_kind}'. Furthermore, it should be the last dimension."
                " Other configurations might lead to suboptimal performance.",
                stacklevel=2,
            )

        # TODO(phimuell): The GPU function might modify the map iteration order.
        #   This is because how it is implemented (promotion and fusion). However,
        #   because of its current state, this should not happen, but we have to look
        #   into it.
        gtx_transformations.gt_gpu_transformation(
            sdfg,
            gpu_block_size=gpu_block_size,
            gpu_launch_bounds=gpu_launch_bounds,
            gpu_launch_factor=gpu_launch_factor,
            gpu_block_size_spec=gpu_block_size_spec,
            gpu_maxnreg=gpu_maxnreg,
            validate=False,
            validate_all=validate_all,
            try_removing_trivial_maps=True,
        )
        if GT4PyAutoOptHook.AfterToGPU in optimization_hooks:
            optimization_hooks[GT4PyAutoOptHook.AfterToGPU](sdfg)  # type: ignore[call-arg]

    return sdfg


def _gt_auto_post_processing(
    sdfg: dace.SDFG,
    gpu: bool,
    make_persistent: bool,
    reuse_transients: bool,
    gpu_memory_pool: bool,
    validate_all: bool,
) -> dace.SDFG:
    """Perform post processing on the SDFG.

    Apply the finishing touch to the optimized SDFG.
    For a full description of the arguments see the `gt_auto_optimize()` function.
    """

    if reuse_transients:
        # TODO(phimuell): Investigate if we should enable it, it may make things
        #   harder for the compiler. Maybe write our own to only consider big
        #   transients and not small ones (~60B)
        transient_reuse = dace.transformation.passes.TransientReuse()
        transient_reuse.apply_pass(sdfg, {})

    # TODO(phimuell): Fix the bug, it uses the tile value and not the stack array value.
    dace_aoptimize.move_small_arrays_to_stack(sdfg)

    if make_persistent and gpu_memory_pool:
        raise ValueError("Cannot set both 'make_persistent' and 'gpu_memory_pool'.")

    if make_persistent:
        device = dace.DeviceType.GPU if gpu else dace.DeviceType.CPU
        gtx_transformations.gt_make_transients_persistent(sdfg=sdfg, device=device)

        if device == dace.DeviceType.GPU:
            # NOTE: For unknown reasons the counterpart of the
            #   `gt_make_transients_persistent()` function in DaCe, resets the
            #   `wcr_nonatomic` property of every memlet, i.e. makes it atomic.
            #   However, it does this only for edges on the top level and on GPU.
            #   For compatibility with DaCe (and until we found out why) the GT4Py
            #   auto optimizer will emulate this behaviour.
            for state in sdfg.states():
                assert isinstance(state, dace.SDFGState)
                for edge in state.edges():
                    edge.data.wcr_nonatomic = False

    if gpu and gpu_memory_pool:
        gtx_transformations.gpu_utils.gt_gpu_apply_mempool(sdfg)

    if validate_all:
        sdfg.validate()

    return sdfg


def gt4py_default_auto_optimizer_map_promotion_checker(
    state: dace.SDFGState,
    sdfg: dace.SDFG,
    first_map_exit: dace_nodes.MapExit,
    second_map_entry: dace_nodes.MapEntry,
    missing_map_parameters: list[str],
) -> bool:
    """Helper function to check if horizontal Map promotion should be performed.

    The function is used by `_gt_auto_process_top_level_maps()` to ensure that
    only horizontal (the function will not check any other promotion) promotions
    that are beneficial will be accepted.

    The implementation checks if the complexity of the first map (the one that
    lacks the horizontal dimension) is not too large, such that inlining is
    beneficial. The rules, which are empirical, are summarized as follows:
    - The number of nodes inside the Map scope is below a certain value (currently
        20 nodes).
    - The Map only contain Tasklets and AccessNodes refering to transient data.
    - There are no WCR edges.

    For arguments see the `MapPromotionCallBack`.
    """

    # We only apply the rule to horizontal promotion. Vertical promotion is known
    #  to be beneficial, especially in conjunction with `LoopBloking`.
    if not any(
        missing_param.endswith("_gtx_horizontal") for missing_param in missing_map_parameters
    ):
        return True

    # NOTE: Inclusion of the exit is important that we get the edges that
    #   actually write the result.
    first_map_entry = state.entry_node(first_map_exit)
    first_map_scope = state.scope_subgraph(first_map_entry, include_entry=True, include_exit=True)

    # If there are more than 20 nodes we assume that the promotion is not
    #  beneficial. This is an arbitrary value.
    if first_map_scope.number_of_nodes() > 22:
        return False

    for node in first_map_scope.nodes():
        if node in (first_map_entry, first_map_exit):
            continue

        # All AccessNodes inside must be transient scalars. Because everything else
        #  indicate (empirical consideration) a more complex computation.
        if isinstance(node, dace_nodes.AccessNode):
            desc = node.desc(sdfg)
            if (not isinstance(desc, dace_data.Scalar)) and (not desc.transient):
                return False
            continue

        # A tasklet is considered to be a cheap computation.
        if isinstance(node, dace_nodes.Tasklet):
            continue

        # Everything else is considered to be a complex computation,
        #  and thus the Map should not be promoted.
        return False

    for edge in first_map_scope.edges():
        # If we have WCR then we have complex computations, most likely
        #   reductions, thus we will refuse the fusing.
        if edge.data.wcr is not None:
            return False

    return True



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import ast
import textwrap
import typing
from typing import Any, Type

import gt4py.eve as eve
from gt4py.next import errors
from gt4py.next.ffront import (
    dialect_ast_enums,
    fbuiltins,
    field_operator_ast as foast,
    source_utils,
    stages as ffront_stages,
)
from gt4py.next.ffront.ast_passes import (
    SingleAssignTargetPass,
    SingleStaticAssignPass,
    StringifyAnnotationsPass,
    UnchainComparesPass,
)
from gt4py.next.ffront.dialect_parser import DialectParser
from gt4py.next.ffront.foast_introspection import StmtReturnKind, deduce_stmt_return_kind
from gt4py.next.ffront.foast_passes.closure_var_folding import ClosureVarFolding
from gt4py.next.ffront.foast_passes.closure_var_type_deduction import ClosureVarTypeDeduction
from gt4py.next.ffront.foast_passes.dead_closure_var_elimination import DeadClosureVarElimination
from gt4py.next.ffront.foast_passes.iterable_unpack import UnpackedAssignPass
from gt4py.next.ffront.foast_passes.type_deduction import FieldOperatorTypeDeduction
from gt4py.next.ffront.stages import (
    ConcreteDSLFieldOperatorDef,
    ConcreteFOASTOperatorDef,
    DSLFieldOperatorDef,
    FOASTOperatorDef,
)
from gt4py.next.otf import toolchain, workflow
from gt4py.next.type_system import type_info, type_specifications as ts, type_translation


def func_to_foast(inp: DSLFieldOperatorDef) -> FOASTOperatorDef:
    """
    Turn a DSL field operator definition into a FOAST operator definition, adding metadata.

    Examples:

        >>> from gt4py import next as gtx
        >>> IDim = gtx.Dimension("I")

        >>> const = gtx.float32(2.0)
        >>> def dsl_operator(a: gtx.Field[[IDim], gtx.float32]) -> gtx.Field[[IDim], gtx.float32]:
        ...     return a * const

        >>> dsl_operator_def = gtx.ffront.stages.DSLFieldOperatorDef(definition=dsl_operator)
        >>> foast_definition = func_to_foast(dsl_operator_def)

        >>> print(foast_definition.foast_node.id)
        dsl_operator

        >>> foast_closure_vars = {k: str(v) for k, v in foast_definition.closure_vars.items()}
        >>> print(foast_closure_vars)
        {'const': '2.0'}
    """
    source_def = source_utils.SourceDefinition.from_function(inp.definition)
    closure_vars = source_utils.get_closure_vars_from_function(inp.definition)
    annotations = typing.get_type_hints(inp.definition)
    foast_definition_node = FieldOperatorParser.apply(source_def, closure_vars, annotations)
    loc = foast_definition_node.location
    operator_attribute_nodes = {
        key: foast.Constant(value=value, type=type_translation.from_value(value), location=loc)
        for key, value in inp.attributes.items()
    }
    untyped_foast_node = inp.node_class(
        id=foast_definition_node.id,
        definition=foast_definition_node,
        location=loc,
        **operator_attribute_nodes,
    )
    foast_node = FieldOperatorTypeDeduction.apply(untyped_foast_node)
    return ffront_stages.FOASTOperatorDef(
        foast_node=foast_node,
        closure_vars=closure_vars,
        grid_type=inp.grid_type,
        attributes=inp.attributes,
        debug=inp.debug,
    )


def func_to_foast_factory(
    cached: bool = True,
) -> workflow.Workflow[DSLFieldOperatorDef, FOASTOperatorDef]:
    """Wrap `func_to_foast` in a chainable and optionally cached workflow step."""
    wf = workflow.make_step(func_to_foast)
    if cached:
        wf = workflow.CachedStep(step=wf, hash_function=ffront_stages.fingerprint_stage)
    return wf


def adapted_func_to_foast_factory(
    **kwargs: Any,
) -> workflow.Workflow[ConcreteDSLFieldOperatorDef, ConcreteFOASTOperatorDef]:
    """Wrap the `func_to_foast step in an adapter to fit into transform toolchains.`"""
    return toolchain.DataOnlyAdapter(func_to_foast_factory(**kwargs))


class FieldOperatorParser(DialectParser[foast.FunctionDefinition]):
    """
    Parse field operator function definition from source code into FOAST.

    Catch any Field Operator specific syntax errors and typing problems.

    Example
    -------
    Parse a function into a Field Operator AST (FOAST), which can
    be lowered into Iterator IR (ITIR)

    >>> from gt4py.next import Field, Dimension
    >>> float64 = float
    >>> IDim = Dimension("IDim")
    >>> def field_op(inp: Field[[IDim], float64]):
    ...     return inp
    >>> foast_tree = FieldOperatorParser.apply_to_function(field_op)
    >>> foast_tree  # doctest: +ELLIPSIS
    FunctionDefinition(id=SymbolName('field_op'), ...)
    >>> foast_tree.params  # doctest: +ELLIPSIS
    [Symbol(id=SymbolName('inp'), type=FieldType(...), ...)]
    >>> foast_tree.body.stmts  # doctest: +ELLIPSIS
    [Return(value=Name(..., id=SymbolRef('inp')))]


    If a syntax error is encountered, it will point to the location in the source code.

    >>> def wrong_syntax(inp: Field[[IDim], int]):
    ...     for i in [1, 2, 3]:  # for is not part of the field operator syntax
    ...         tmp = inp
    ...     return tmp
    >>>
    >>> try:  # doctest: +ELLIPSIS
    ...     FieldOperatorParser.apply_to_function(wrong_syntax)
    ... except errors.DSLError as err:
    ...     print(
    ...         f"Error at [{err.location.line}, {err.location.column}] in {err.location.filename})"
    ...     )
    Error at [2, 5] in ...func_to_foast.FieldOperatorParser[...]>)
    """

    @classmethod
    def _preprocess_definition_ast(cls, ast: ast.AST) -> ast.AST:
        ast = StringifyAnnotationsPass.apply(ast)
        ast = SingleStaticAssignPass.apply(ast)
        ast = SingleAssignTargetPass.apply(ast)
        return ast

    @classmethod
    def _postprocess_dialect_ast(
        cls,
        foast_node: foast.FunctionDefinition | foast.FieldOperator,
        closure_vars: dict[str, Any],
        annotations: dict[str, Any],
    ) -> foast.FunctionDefinition:
        foast_node = ClosureVarFolding.apply(foast_node, closure_vars)
        foast_node = DeadClosureVarElimination.apply(foast_node)
        foast_node = ClosureVarTypeDeduction.apply(foast_node, closure_vars)
        foast_node = FieldOperatorTypeDeduction.apply(foast_node)
        foast_node = UnpackedAssignPass.apply(foast_node)

        # check deduced matches annotated return type
        if "return" in annotations:
            annotated_return_type = type_translation.from_type_hint(annotations["return"])
            # TODO(tehrengruber): use `type_info.return_type` when the type of the
            #  arguments becomes available here
            if annotated_return_type != foast_node.type.returns:  # type: ignore[union-attr] # revisit when `type_info.return_type` is implemented
                raise errors.DSLError(
                    foast_node.location,
                    "Annotated return type does not match deduced return type: annotation is "
                    f"'{annotated_return_type}'"  # type: ignore[union-attr] # revisit when 'type_info.return_type' is implemented
                    f", got '{foast_node.type.returns}'.",
                )
        return foast_node

    def visit_FunctionDef(self, node: ast.FunctionDef, **kwargs: Any) -> foast.FunctionDefinition:
        loc = self.get_location(node)
        closure_var_symbols: list[foast.Symbol] = []
        for name in self.closure_vars.keys():
            try:
                type_ = type_translation.from_value(self.closure_vars[name])
                closure_var_symbols.append(
                    foast.Symbol(
                        id=name,
                        type=type_,
                        namespace=dialect_ast_enums.Namespace.CLOSURE,
                        location=self.get_location(node),
                    )
                )
            except ValueError as e:
                raise errors.DSLTypeError(
                    loc,
                    f"Unexpected object '{name}' of type '{type(self.closure_vars[name])}' encountered.",
                ) from e

        new_body = self._visit_stmts(node.body, self.get_location(node), **kwargs)

        if deduce_stmt_return_kind(new_body) == StmtReturnKind.NO_RETURN:
            raise errors.DSLError(loc, "'Function' is expected to return a value.")

        return foast.FunctionDefinition(
            id=node.name,
            params=self.visit(node.args, **kwargs),
            body=new_body,
            closure_vars=closure_var_symbols,
            location=loc,
        )

    def visit_arguments(self, node: ast.arguments) -> list[foast.DataSymbol]:
        return [self.visit_arg(arg) for arg in node.args]

    def visit_arg(self, node: ast.arg) -> foast.DataSymbol:
        loc = self.get_location(node)
        if (annotation := self.annotations.get(node.arg, None)) is None:
            raise errors.MissingParameterAnnotationError(loc, node.arg)
        new_type = type_translation.from_type_hint(annotation)
        if not isinstance(new_type, ts.DataType):
            raise errors.InvalidParameterAnnotationError(loc, node.arg, new_type)
        return foast.DataSymbol(id=node.arg, location=loc, type=new_type)

    def visit_Assign(
        self, node: ast.Assign, **kwargs: Any
    ) -> foast.Assign | foast.TupleTargetAssign:
        target = node.targets[0]  # there is only one element after assignment passes

        if isinstance(target, ast.Tuple):
            new_targets: list[
                foast.FieldSymbol | foast.TupleSymbol | foast.ScalarSymbol | foast.Starred
            ] = []

            for elt in target.elts:
                if isinstance(elt, ast.Starred):
                    new_targets.append(
                        foast.Starred(
                            id=foast.DataSymbol(
                                id=self.visit(elt.value).id,
                                location=self.get_location(elt),
                                type=ts.DeferredType(constraint=ts.DataType),
                            ),
                            location=self.get_location(elt),
                            type=ts.DeferredType(constraint=ts.DataType),
                        )
                    )
                else:
                    new_targets.append(
                        foast.DataSymbol(
                            id=self.visit(elt).id,
                            location=self.get_location(elt),
                            type=ts.DeferredType(constraint=ts.DataType),
                        )
                    )

            return foast.TupleTargetAssign(
                targets=new_targets, value=self.visit(node.value), location=self.get_location(node)
            )

        if not isinstance(target, ast.Name):
            raise errors.DSLError(self.get_location(node), "Can only assign to names.")
        new_value = self.visit(node.value)
        constraint_type: Type[ts.DataType] = ts.DataType
        if isinstance(new_value, foast.TupleExpr):
            constraint_type = ts.TupleType
        elif (
            type_info.is_concrete(new_value.type)
            and type_info.type_class(new_value.type) is ts.ScalarType
        ):
            constraint_type = ts.ScalarType
        return foast.Assign(
            target=foast.DataSymbol(
                id=target.id,
                location=self.get_location(target),
                type=ts.DeferredType(constraint=constraint_type),
            ),
            value=new_value,
            location=self.get_location(node),
        )

    def visit_AnnAssign(self, node: ast.AnnAssign, **kwargs: Any) -> foast.Assign:
        if not isinstance(node.target, ast.Name):
            raise errors.DSLError(self.get_location(node), "Can only assign to names.")

        if node.annotation is not None:
            assert isinstance(node.annotation, ast.Constant) and isinstance(
                node.annotation.value, str
            ), "Annotations should be ast.Constant(string). Use StringifyAnnotationsPass"

            context = {**fbuiltins.BUILTINS, **self.closure_vars}
            annotation = eval(node.annotation.value, context)
            target_type = type_translation.from_type_hint(annotation, globalns=context)
        else:
            target_type = ts.DeferredType()

        return foast.Assign(
            target=foast.Symbol[ts.FieldType](
                id=node.target.id, location=self.get_location(node.target), type=target_type
            ),
            value=self.visit(node.value) if node.value else None,
            location=self.get_location(node),
        )

    def visit_Subscript(self, node: ast.Subscript, **kwargs: Any) -> foast.Subscript:
        return foast.Subscript(
            value=self.visit(node.value),
            index=self.visit(node.slice),
            location=self.get_location(node),
        )

    def visit_Attribute(self, node: ast.Attribute) -> Any:
        return foast.Attribute(
            value=self.visit(node.value), attr=node.attr, location=self.get_location(node)
        )

    def visit_Tuple(self, node: ast.Tuple, **kwargs: Any) -> foast.TupleExpr:
        return foast.TupleExpr(
            elts=[self.visit(item) for item in node.elts], location=self.get_location(node)
        )

    def visit_Return(self, node: ast.Return, **kwargs: Any) -> foast.Return:
        loc = self.get_location(node)
        if not node.value:
            raise errors.DSLError(loc, "Must return a value, not None")
        return foast.Return(value=self.visit(node.value), location=loc)

    def visit_Expr(self, node: ast.Expr) -> foast.Expr:
        return self.visit(node.value)

    def visit_Name(self, node: ast.Name, **kwargs: Any) -> foast.Name:
        return foast.Name(id=node.id, location=self.get_location(node))

    def visit_UnaryOp(self, node: ast.UnaryOp, **kwargs: Any) -> foast.UnaryOp:
        return foast.UnaryOp(
            op=self.visit(node.op),
            operand=self.visit(node.operand),
            location=self.get_location(node),
        )

    def visit_UAdd(self, node: ast.UAdd, **kwargs: Any) -> dialect_ast_enums.UnaryOperator:
        return dialect_ast_enums.UnaryOperator.UADD

    def visit_USub(self, node: ast.USub, **kwargs: Any) -> dialect_ast_enums.UnaryOperator:
        return dialect_ast_enums.UnaryOperator.USUB

    def visit_Not(self, node: ast.Not, **kwargs: Any) -> dialect_ast_enums.UnaryOperator:
        return dialect_ast_enums.UnaryOperator.NOT

    def visit_Invert(self, node: ast.Invert, **kwargs: Any) -> dialect_ast_enums.UnaryOperator:
        return dialect_ast_enums.UnaryOperator.INVERT

    def visit_BinOp(self, node: ast.BinOp, **kwargs: Any) -> foast.BinOp:
        return foast.BinOp(
            op=self.visit(node.op),
            left=self.visit(node.left),
            right=self.visit(node.right),
            location=self.get_location(node),
        )

    def visit_Add(self, node: ast.Add, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.ADD

    def visit_Sub(self, node: ast.Sub, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.SUB

    def visit_Mult(self, node: ast.Mult, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.MULT

    def visit_Div(self, node: ast.Div, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.DIV

    def visit_FloorDiv(self, node: ast.FloorDiv, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.FLOOR_DIV

    def visit_Pow(self, node: ast.Pow, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.POW

    def visit_Mod(self, node: ast.Mod, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.MOD

    def visit_BitAnd(self, node: ast.BitAnd, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.BIT_AND

    def visit_BitOr(self, node: ast.BitOr, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.BIT_OR

    def visit_BitXor(self, node: ast.BitXor, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.BIT_XOR

    def visit_BoolOp(self, node: ast.BoolOp, **kwargs: Any) -> None:
        raise errors.UnsupportedPythonFeatureError(
            self.get_location(node), "logical operators `and`, `or`"
        )

    def visit_IfExp(self, node: ast.IfExp, **kwargs: Any) -> foast.TernaryExpr:
        return foast.TernaryExpr(
            condition=self.visit(node.test),
            true_expr=self.visit(node.body),
            false_expr=self.visit(node.orelse),
            location=self.get_location(node),
            type=ts.DeferredType(constraint=ts.DataType),
        )

    def visit_If(self, node: ast.If, **kwargs: Any) -> foast.IfStmt:
        loc = self.get_location(node)
        return foast.IfStmt(
            condition=self.visit(node.test, **kwargs),
            true_branch=self._visit_stmts(node.body, loc, **kwargs),
            false_branch=self._visit_stmts(node.orelse, loc, **kwargs),
            location=loc,
        )

    def _visit_stmts(
        self, stmts: list[ast.stmt], location: eve.SourceLocation, **kwargs: Any
    ) -> foast.BlockStmt:
        return foast.BlockStmt(
            stmts=[self.visit(el, **kwargs) for el in stmts if not isinstance(el, ast.Pass)],
            location=location,
        )

    def visit_Compare(self, node: ast.Compare, **kwargs: Any) -> foast.Compare:
        loc = self.get_location(node)

        if len(node.ops) != 1 or len(node.comparators) != 1:
            refactored = UnchainComparesPass.apply(node)
            raise errors.DSLError(
                loc,
                textwrap.dedent(
                    f"""
                    Comparison chains are not allowed. Please replace
                        {ast.unparse(node)}
                    by
                        {ast.unparse(refactored)}
                    """,
                ),
            )
        return foast.Compare(
            op=self.visit(node.ops[0]),
            left=self.visit(node.left),
            right=self.visit(node.comparators[0]),
            location=loc,
        )

    def visit_Gt(self, node: ast.Gt, **kwargs: Any) -> foast.CompareOperator:
        return foast.CompareOperator.GT

    def visit_Lt(self, node: ast.Lt, **kwargs: Any) -> foast.CompareOperator:
        return foast.CompareOperator.LT

    def visit_Eq(self, node: ast.Eq, **kwargs: Any) -> foast.CompareOperator:
        return foast.CompareOperator.EQ

    def visit_LtE(self, node: ast.LtE, **kwargs: Any) -> foast.CompareOperator:
        return foast.CompareOperator.LTE

    def visit_GtE(self, node: ast.GtE, **kwargs: Any) -> foast.CompareOperator:
        return foast.CompareOperator.GTE

    def visit_NotEq(self, node: ast.NotEq, **kwargs: Any) -> foast.CompareOperator:
        return foast.CompareOperator.NOTEQ

    def _verify_builtin_type_constructor(self, node: ast.Call) -> None:
        if len(node.args) > 0:
            arg = node.args[0]
            if not (
                isinstance(arg, ast.Constant)
                or (isinstance(arg, ast.UnaryOp) and isinstance(arg.operand, ast.Constant))
            ):
                raise errors.DSLError(
                    self.get_location(node),
                    f"'{self._func_name(node)}()' only takes literal arguments.",
                )

    def _func_name(self, node: ast.Call) -> str:
        return node.func.id  # type: ignore[attr-defined] # We want this to fail if the attribute does not exist unexpectedly.

    def visit_Call(self, node: ast.Call, **kwargs: Any) -> foast.Call:
        # TODO(tehrengruber): is this still needed or redundant with the checks in type deduction?
        if isinstance(node.func, ast.Name):
            func_name = self._func_name(node)
            if func_name in fbuiltins.TYPE_BUILTIN_NAMES:
                self._verify_builtin_type_constructor(node)

        return foast.Call(
            func=self.visit(node.func, **kwargs),
            args=[self.visit(arg, **kwargs) for arg in node.args],
            kwargs={keyword.arg: self.visit(keyword.value, **kwargs) for keyword in node.keywords},
            location=self.get_location(node),
        )

    def visit_Constant(self, node: ast.Constant, **kwargs: Any) -> foast.Constant:
        loc = self.get_location(node)
        try:
            type_ = type_translation.from_value(node.value)
        except ValueError:
            raise errors.DSLError(
                loc, f"Constants of type {type(node.value)} are not permitted."
            ) from None

        return foast.Constant(value=node.value, location=loc, type=type_)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Mapping, Optional

import dace
from dace import properties as dace_properties, transformation as dace_transformation
from dace.sdfg import nodes as dace_nodes
from dace.transformation.passes import analysis as dace_analysis

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


def gt_eliminate_dead_dataflow(
    sdfg: dace.SDFG,
    run_simplify: bool,
    validate: bool = True,
    validate_all: bool = False,
) -> int:
    """Performs dead dataflow elimination on the `sdfg`.

    The function will run GT4Py's dead datflow elimination transformation and if
    requested simplify afterwards. It is recommended to use this function instead
    as it will also run `FindSingleUseData` before.
    The function will return the number maps and memlets that have been removed,
    the output of `gt_simplify()`, if any, is ignored.

    Args:
        sdfg: The SDFG to process.
        run_simplify: Run `gt_simplify()` after the dead datflow elimination.
        validate: Perform validation.
        validate_all: Perform extensive validation.

    Todo:
        Implement a better way of applying the `DeadMemletElimination` transformation.
    """
    find_single_use_data = dace_analysis.FindSingleUseData()
    single_use_data = find_single_use_data.apply_pass(sdfg, None)

    ret = sdfg.apply_transformations_once_everywhere(
        DeadMapElimination(single_use_data=single_use_data),
        validate=False,
        validate_all=validate_all,
    )
    ret += gt_dead_memlet_elimination(
        sdfg=sdfg,
        single_use_data=single_use_data,
        validate=False,
        validate_all=validate_all,
    )

    if run_simplify:
        gtx_transformations.gt_simplify(
            sdfg=sdfg,
            validate=False,
            validate_all=validate_all,
        )
    if validate and (not validate_all):
        sdfg.validate()

    return ret


@dace_properties.make_properties
class DeadMapElimination(dace_transformation.SingleStateTransformation):
    """The transformation removes all Maps that are no-ops.

    The transformation matches any MapEntry, at the top level and evaluates the ranges,
    if it concludes that the map never execute, because at least one dimension as
    zero or less iteration steps, it will remove the entire Map.

    The function will also remove the nodes that have become isolated. If the optional
    `single_use_data` was passed at construction, the transformation will also remove
    the data from the registry, i.e. `SDFG.arrays`.
    """

    map_entry = dace_transformation.PatternNode(dace_nodes.MapEntry)
    _single_use_data: Optional[dict[dace.SDFG, set[str]]]

    def __init__(
        self,
        single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self._single_use_data = None
        if single_use_data is not None:
            self._single_use_data = single_use_data
        super().__init__(*args, **kwargs)

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.map_entry)]

    def can_be_applied(
        self,
        graph: dace.SDFGState,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        map_entry: dace_nodes.MapEntry = self.map_entry
        map_range: dace.subsets.Range = map_entry.map.range
        scope_dict = graph.scope_dict()

        # The map must be on the top.
        if scope_dict[map_entry] is not None:
            return False

        # Look if there is a dimension with zero or "less" iteration steps.
        if not any((mr <= 0) == True for mr in map_range.size()):  # noqa: E712 [true-false-comparison]  # SymPy fuzzy bools.
            return False

        # For ease of implementation we also require that the Map scope is only
        #  adjacent to AccessNodes.
        if not all(
            isinstance(iedge.src, dace_nodes.AccessNode) for iedge in graph.in_edges(map_entry)
        ):
            return False

        map_exit: dace_nodes.MapExit = graph.exit_node(map_entry)
        if not all(
            isinstance(oedge.dst, dace_nodes.AccessNode) for oedge in graph.out_edges(map_exit)
        ):
            return False

        return True

    def apply(
        self,
        graph: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> None:
        gt_remove_map(
            sdfg=sdfg,
            state=graph,
            map_entry=self.map_entry,
            remove_unused_data=(self._single_use_data is not None),
            single_use_data=self._single_use_data,
        )


def gt_remove_map(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    map_entry: dace_nodes.MapEntry,
    remove_unused_data: bool = False,
    single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
) -> int:
    """Remove the Map denoted by `map_entry` _unconditionally_.

    The function can only operate on top level Maps, nested Maps are not supported.
    Furthermore, adjacent nodes that have become isolated by the Map removal will
    also be removed.
    If `remove_unused_data` is specified, then the function will also remove data
    descriptors that were referenced by the Map and are no longer used.

    Note, it is the user's responsibility to ensure that the removal of the Map is
    valid as no checks are performed.

    The function returns the number of nodes that were removed.

    Args:
        sdfg: The SDFG on which we operate.
        state: The state that contains the Map.
        map_entry: The entry node of the Map that is removed.
        remove_unused_data: Remove data descriptors that were referred inside the
            Map and are no longer needed.
        single_use_data: Single use data, see `dace.FindSingleUseData()`. Only
            needed if `remove_unused_data` is set.
    """

    if state.scope_dict()[map_entry] is not None:
        raise ValueError("Can only remove Map-Scopes on the top level.")

    if single_use_data is None and remove_unused_data:
        # We have to do it here, because data that is not referred anywhere in the
        #  SDFG is classified as "not single use data".
        find_single_use_data = dace_analysis.FindSingleUseData()
        single_use_data = find_single_use_data.apply_pass(sdfg, None)

    map_exit: dace_nodes.MapExit = state.exit_node(map_entry)
    map_scope = state.scope_subgraph(map_entry, include_exit=True, include_entry=True)

    # Now find all notes that are producer or consumer of the Map, after we removed
    #  the nodes of the Maps we need to check if they have become isolated.
    # NOTE: Needs to be a `set` to handle multiple connections with the MapEntry node.
    adjacent_nodes: set[dace_nodes.AccessNode] = {iedge.src for iedge in state.in_edges(map_entry)}
    adjacent_nodes.update(oedge.dst for oedge in state.out_edges(map_exit))

    if not all(isinstance(ac, dace_nodes.AccessNode) for ac in adjacent_nodes):
        raise ValueError(
            f"Expected that all adjacent nodes of '{map_entry.map.label}' are AccessNode"
        )

    # Find all the data that is accessed inside the Map scope.
    map_scope_datas: set[str] = (
        {ac.data for ac in map_scope.data_nodes()} if remove_unused_data else set()
    )

    # Now remove the nodes of the scope.
    removed_number_nodes = map_scope.number_of_nodes()
    state.remove_nodes_from(map_scope.nodes())

    # Now check for potentially isolated nodes.
    #  Process them in deterministic order.
    for adjacent_node in sorted(adjacent_nodes, key=lambda ac: ac.data):
        if state.degree(adjacent_node) == 0:
            map_scope_datas.add(adjacent_node.data)
            state.remove_node(adjacent_node)
            removed_number_nodes += 1

    if not remove_unused_data:
        return removed_number_nodes

    # If requested remove the data that is no longer referenced.
    single_use_data_sdfg = single_use_data[sdfg]  # type: ignore[index]  # Guaranteed to be not None.
    for map_scope_data in map_scope_datas:
        if not sdfg.arrays[map_scope_data].transient:
            continue
        if map_scope_data in single_use_data_sdfg:
            sdfg.remove_data(map_scope_data, validate=False)

    return removed_number_nodes


def gt_dead_memlet_elimination(
    sdfg: dace.SDFG,
    single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
    validate: bool = True,
    validate_all: bool = False,
) -> int:
    """Removes Memlets that do not carry any dataflow.

    The function will recursively examine the outgoing Memlets of all AccessNodes.
    If it finds a Memlet that does not carry any data, the Memlet will be removed.
    However, the function will only consider connections between two AccessNodes.
    If the removing of the Memlet leads to isolated nodes, these will be removed.
    Furthermore, if `single_use_data` was passed the function will also remove
    the data descriptors that have become obsolete from the registry.

    Args:
        sdfg: The SDFG that is processed.
        single_use_data: List of data that is only used at a single location.
            If passed the function will also clean the registry.

    Return:
        The number of Memlets that were removed.
    """
    ret = 0
    for rsdfg in sdfg.all_sdfgs_recursive():
        ret += _gt_dead_memlet_elimination_sdfg(sdfg=rsdfg, single_use_data=single_use_data)

        if validate_all:
            rsdfg.validate()

    if validate and (not validate_all):
        sdfg.validate()

    return ret


def _gt_dead_memlet_elimination_sdfg(
    sdfg: dace.SDFG,
    single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
) -> int:
    """Removes Memlets that do not carry any dataflow.

    The same as `gt_dead_memlet_elimination` but only operates on a single level.
    """

    def _find_candidates_for(
        access_node: dace_nodes.AccessNode,
        state: dace.SDFGState,
    ) -> list[dace.sdfg.state.MultiConnectorEdge]:
        candidates: list[dace.sdfg.MultiConnectorEdge] = []
        for oedge in state.out_edges(access_node):
            dst: dace_nodes.Node = oedge.dst
            memlet: dace.Memlet = oedge.data
            if not isinstance(dst, dace_nodes.AccessNode):
                continue
            if memlet.is_empty():
                continue
            subset = memlet.subset if memlet.subset is not None else memlet.other_subset
            assert subset is not None, "Failed to identify the subset."
            if not any((ss <= 0) == True for ss in subset.size()):  # noqa: E712 [true-false-comparison]  # SymPy fuzzy bools.
                continue
            candidates.append(oedge)
        return candidates

    def _process_access_node(
        access_node: dace_nodes.AccessNode,
        state: dace.SDFGState,
    ) -> tuple[int, set[dace_nodes.AccessNode]]:
        removed_access_nodes: set[dace_nodes.AccessNode] = set()
        removed_memlets: int = 0
        for oedge in _find_candidates_for(access_node, state):
            other_node: dace_nodes.AccessNode = oedge.dst
            state.remove_edge(oedge)
            removed_memlets += 1
            # TODO(phimuell): Handle the case if there are only empty Memlets.
            if state.degree(other_node) == 0:
                state.remove_node(other_node)
                removed_access_nodes.add(other_node)
        # TODO(phimuell): Handle the case if there are only empty Memlets.
        if state.degree(access_node) == 0:
            state.remove_node(access_node)
            removed_access_nodes.add(access_node)
        return removed_memlets, removed_access_nodes

    # Now go through all states and process the data nodes, also keep track of all
    removed_memlets: int = 0
    for state in sdfg.states():
        # Keep track of all AccessNodes that have been removed, for the later cleaning
        #  and to prevent `NondeNotFoundError`s in DaCe.
        removed_access_nodes: set[dace_nodes.AccessNode] = set()
        for dnode in list(state.data_nodes()):
            if dnode in removed_access_nodes:
                continue
            this_removed_memlets, this_removed_access_nodes = _process_access_node(
                access_node=dnode,
                state=state,
            )
            removed_memlets += this_removed_memlets
            removed_access_nodes.update(this_removed_access_nodes)

        # Now remove the data that is no longer needed from the registry.
        if single_use_data is not None and len(removed_access_nodes) != 0:
            sdfg_single_use_data = single_use_data[sdfg]
            removed_data_names: set[str] = {ac.data for ac in removed_access_nodes}
            arrays: Mapping[str, dace.data.Data] = sdfg.arrays
            for removed_data in removed_data_names:
                if not arrays[removed_data].transient:
                    continue
                if removed_data in sdfg_single_use_data:
                    sdfg.remove_data(removed_data, validate=False)

    return removed_memlets



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Implements the lowering of concat_where operator.

This builtin translator implements the `PrimitiveTranslator` protocol as other
translators in `gtir_to_sdfg_primitives` module.
"""

from __future__ import annotations

from typing import Sequence

import dace
from dace import nodes as dace_nodes, subsets as dace_subsets

from gt4py.next import common as gtx_common
from gt4py.next.iterator import ir as gtir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    domain_utils,
    ir_makers as im,
)
from gt4py.next.program_processors.runners.dace import sdfg_args as gtx_dace_args
from gt4py.next.program_processors.runners.dace.lowering import (
    gtir_domain,
    gtir_to_sdfg,
    gtir_to_sdfg_types,
    gtir_to_sdfg_utils,
)
from gt4py.next.type_system import type_specifications as ts


def _translate_concat_where_branch(
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    source_expr: gtir.Expr,
    is_lower: bool,
    concat_dim: gtx_common.Dimension,
    concat_dim_bound_expr: gtir.Expr,
    output_domain: domain_utils.SymbolicDomain,
    output_type: ts.FieldType,
    output_desc: dace.data.Array,
    output_node: dace_nodes.AccessNode,
    output_origin: Sequence[dace.symbolic.SymbolicType],
) -> None:
    """
    Translate one of the two branches of a 'concat_where' expression.

    The 'concat_where' expression requires two input fields, one is written on the
    lower part of the result domain, with respect to the domain boundary extracted
    from the first argument; the other input field is written on the upper domain.
    The handling of the two branches is similar, there is only one small difference
    in case the input field does not contain the 'concat_where' dimension, which
    is the case of scalar values or horizontal slice fields. In this case, we use
    the 'is_lower' argument to specify which of the two branches this function
    should target.

    The _regular_ case is `concat_where` applied to two fields with same domain
    as the result field, for example two 'IJK'-fields:
    ```python
    @gtx.field_operator
    def testee(a: cases.IJKField, b: cases.IJKField) -> cases.IJKField:
        return concat_where(KDim < 10, a, b)
    ```

    In case of the arguments is a scalar value, we broadcast it on all dimensions
    of the result field, for example:
    ```python
    @gtx.field_operator
    def testee(a: np.int32, b: IJKField) -> IJKField:
        return concat_where(KDim == 0, a, b)
    ```
    Similarly, we broadcast a field defined as a slice of the output domain, i.e.
    a field with a smaller number of dimensions than the result field.
    Consider for example the following field operator, where the 'boundary' IJ-field
    is used for the vertical boundary (`KDim == 0`):
    ```python
    @gtx.field_operator
    def testee(interior: cases.IJKField, boundary: cases.IJField) -> cases.IJKField:
        return concat_where(KDim == 0, boundary, interior)
    ```

    Args:
        ctx: The SDFG context in which to lower the `concat_where` branch expression.
        sdfg_builder: The visitor object to build the SDFG.
        source_expr: The expression producing the input field for this branch.
        is_lower: Flag to specify whether the input expression is for the lower or upper domain.
        concat_dim: The dimension along which to apply `concat_where` on the input fields.
        concat_dim_bound_expr: The expression of the `concat_where` domain condition.
        output_domain: The domain of the result field.
        output_type: The GT4Py type descriptor of the result field.
        output_desc: The DaCe array descriptor of the result field.
        output_node: The SDFG access node for writing the result field.
        output_origin: The origin of the result field, derived from `output_domain`,
            to be used in the write memlet to `output_node`.
    """
    assert isinstance(source_expr.type, (ts.FieldType, ts.ScalarType))

    source_domain = source_expr.annex.domain
    if isinstance(source_expr.type, ts.ScalarType) or len(source_expr.type.dims) < len(
        output_type.dims
    ):
        # We promote the input expression to a field defined on the output domain,
        # refer to the function documentation for examples of such field operators.
        if concat_dim not in source_domain.ranges:
            source_domain.ranges[concat_dim] = (
                domain_utils.SymbolicRange(
                    start=output_domain.ranges[concat_dim].start,
                    stop=concat_dim_bound_expr,
                )
                if is_lower
                else domain_utils.SymbolicRange(
                    start=concat_dim_bound_expr,
                    stop=output_domain.ranges[concat_dim].stop,
                )
            )
        source_domain = domain_utils.promote_domain(source_domain, output_type.dims)
        source_domain = domain_utils.domain_intersection(source_domain, output_domain)

        # Use a 'deref' field operator to broadcast the input expression on the target domain.
        bcast_expr = im.as_fieldop("deref", source_domain.as_expr())(source_expr)
        bcast_expr.type = output_type

        source = sdfg_builder.visit(bcast_expr, ctx=ctx)
    else:
        # The input field is defined on all dimensions of the result field.
        source = sdfg_builder.visit(source_expr, ctx=ctx)

    assert source.gt_type == output_type
    source_domain_range = source_domain.ranges[concat_dim]
    source_range_0 = gtir_to_sdfg_utils.get_symbolic(source_domain_range.start)
    source_range_1 = gtir_to_sdfg_utils.get_symbolic(
        im.maximum(source_domain_range.start, source_domain_range.stop)
    )
    source_range_size = source_range_1 - source_range_0

    if isinstance(output_type.dtype, ts.ScalarType):
        all_dims = gtx_common.order_dimensions(output_type.dims)
    else:
        assert output_type.dtype.offset_type
        all_dims = gtx_common.order_dimensions([*output_type.dims, output_type.dtype.offset_type])

    source_subset = []
    output_subset = []
    for dim, size, src_origin, dst_origin in zip(
        all_dims,
        output_desc.shape,
        source.origin,
        output_origin,
        strict=True,
    ):
        if dim == concat_dim:
            # Write only the subset corresponding to the range of lower or upper branch.
            source_subset.append(
                (
                    source_range_0 - src_origin,
                    source_range_1 - src_origin - 1,
                    1,
                )
            )
            output_subset.append(
                (
                    source_range_0 - dst_origin,
                    source_range_0 - dst_origin + source_range_size - 1,
                    1,
                )
            )
        else:
            # Write the full subset which covers the array size in this dimension.
            source_subset.append(
                (
                    dst_origin - src_origin,
                    dst_origin - src_origin + size - 1,
                    1,
                )
            )
            output_subset.append((0, size - 1, 1))

    ctx.state.add_nedge(
        source.dc_node,
        output_node,
        dace.Memlet(
            data=source.dc_node.data,
            subset=dace_subsets.Range(source_subset),
            other_subset=dace_subsets.Range(output_subset),
        ),
    )


def translate_concat_where(
    node: gtir.Node,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
) -> gtir_to_sdfg_types.FieldopResult:
    """
    Lowers a `concat_where` expression to a dataflow where two memlets write
    disjoint subsets, for the lower and upper domain, on one data access node.

    Implements the `PrimitiveTranslator` protocol.
    """
    assert cpm.is_call_to(node, "concat_where")
    assert len(node.args) == 3
    assert isinstance(node.type, (ts.FieldType, ts.TupleType))

    if isinstance(node.type, ts.TupleType):
        raise NotImplementedError("Unexpected 'concat_where' with tuple output in SDFG lowering.")

    # First argument is a domain expression that defines the mask of the true branch:
    # we extract the dimension along which we need to concatenate the field arguments,
    # and determine whether the true branch argument should be on the lower or upper
    # range with respect to the boundary value.
    mask_domain = domain_utils.SymbolicDomain.from_expr(node.args[0])
    if len(mask_domain.ranges) != 1:
        raise NotImplementedError("Expected `concat_where` along single axis.")

    concat_dim = next(iter(mask_domain.ranges.keys()))

    # Expect unbound range in the concat domain expression on range start or end:
    #  - If the domain expression is unbound on range start (negative infinite),
    #    the expression on the true branch is to be considered the input for the
    #    lower domain.
    #  - Vice versa, if the domain expression is unbound on range stop (positive
    #    infinite), the true expression represents the input for the upper domain.
    infinity_literals = (gtir.InfinityLiteral.POSITIVE, gtir.InfinityLiteral.NEGATIVE)
    if mask_domain.ranges[concat_dim].start in infinity_literals:
        bound_expr = mask_domain.ranges[concat_dim].stop
        lower_expr, upper_expr = node.args[1:]
    elif mask_domain.ranges[concat_dim].stop in infinity_literals:
        bound_expr = mask_domain.ranges[concat_dim].start
        upper_expr, lower_expr = node.args[1:]
    else:
        raise ValueError(f"Unexpected concat mask {mask_domain} with finite domain.")

    # We use the concat domain, stored in the annex, as the domain of output field.
    output_domain = gtir_domain.get_field_domain(node.annex.domain)
    output_dims, output_origin, output_shape = gtir_domain.get_field_layout(output_domain)
    assert output_dims == node.type.dims

    if isinstance(node.type.dtype, ts.ScalarType):
        dtype = gtx_dace_args.as_dace_type(node.type.dtype)
    else:
        # TODO(edopao): Refactor allocation of fields with local dimension and enable this.
        raise NotImplementedError("'concat_where' with list output is not supported")

    output, output_desc = sdfg_builder.add_temp_array(ctx.sdfg, output_shape, dtype)
    output_node = ctx.state.add_access(output)

    # Translate the input expression on the lower domain.
    _translate_concat_where_branch(
        ctx=ctx,
        sdfg_builder=sdfg_builder,
        source_expr=lower_expr,
        is_lower=True,
        concat_dim=concat_dim,
        concat_dim_bound_expr=bound_expr,
        output_domain=node.annex.domain,
        output_type=node.type,
        output_desc=output_desc,
        output_node=output_node,
        output_origin=output_origin,
    )

    # Translate the input expression on the upper domain.
    _translate_concat_where_branch(
        ctx=ctx,
        sdfg_builder=sdfg_builder,
        source_expr=upper_expr,
        is_lower=False,
        concat_dim=concat_dim,
        concat_dim_bound_expr=bound_expr,
        output_domain=node.annex.domain,
        output_type=node.type,
        output_desc=output_desc,
        output_node=output_node,
        output_origin=output_origin,
    )

    return gtir_to_sdfg_types.FieldopData(output_node, node.type, origin=tuple(output_origin))
